# NOTE: This import allows for forward references in type annotations.
from __future__ import annotations


__all__ = [
    'Level', 'SimpleLevel', 'DerivedLevel', 'ElseLevel',
    'Factor', 'SimpleFactor', 'DerivedFactor',
    'WithinTrial', 'Transition', 'Window'
]


from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from itertools import product, chain
from random import randint
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

from sweetpea._internal.iter import chunk_dict
from sweetpea._internal.beforestart import BeforeStart


###############################################################################
##
## Levels
##


@dataclass
class Level:
    """A discrete value that a :class:`.Factor` can hold.

    For more on levels, consult :ref:`the guide <guide_factorial_levels>`.
    """

    #: The name of the level.
    name: str

    # TODO: There is probably a way to make this smoother. As it stands,
    #       factors are responsible for registering themselves with their
    #       levels. Perhaps factors could be the only way to create levels?
    #: The factor to which this level belongs.
    factor: Factor = field(init=False, repr=False)

    # NOTE: Because of the way dataclasses work, a base class (e.g., `Level`)
    #       cannot provide an attribute with a default value if a subclass
    #       (e.g., `DerivedLevel`) provides additional non-default values. This
    #       means the base class cannot provide a `weight` field with a default
    #       value of `1`, which is what we want to achieve. (Additionally,
    #       we'd prefer the `weight` to always be the last field of a level's
    #       initialization parameters.)
    #           The solution is to have the base class implement a hidden
    #       attribute, and provide access to it via a `@property`. Then, the
    #       child classes must add a `weight` `InitVar` field with a default
    #       value after their other fields, and use the `__post_init__` to set
    #       `_weight`. This looks something like:
    #
    #           @dataclass
    #           class NewLevel(Level):
    #               ...
    #               weight: InitVar[int] = 1
    #
    #               def __post_init__(self, ..., weight: int):
    #                   ...
    #                   self._weight = weight
    #                   self.weight = weight
    #                   ...
    _weight: int = field(init=False)

    _synthesized: bool = field(init=False)

    def __new__(cls, *_, **__):
        if cls == Level:
            return super().__new__(SimpleLevel)
        else:
            return super().__new__(cls)

    def __post_init__(self):
        self._synthesized = False

    def __str__(self) -> str:
        return f"Level<{self.name}>"

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        # only equal when it's the same object:
        return id(self) == id(other)

    @property
    def weight(self) -> int:
        """The level's weight."""
        return self._weight

    def uses_factor(self, factor):
        return False

    def get_factor(self) -> Factor:
        if hasattr(self, 'factor'):
            return cast(Factor, cast(Any, self).factor)
        else:
            raise Exception("no factor present")

@dataclass(eq=False)
class SimpleLevel(Level):
    """A simple :class:`.Level`, which only has a name.

    For example, in a simple :ref:`Stroop experiment
    <guide_factorial_example>`, a ``color`` :class:`.Factor` may consist of a
    few simple levels such as ``red``, ``blue``, and ``green``.

    :param name:
        The name of the level.
    """

    weight: InitVar[int] = 1

    # NOTE: The __post_init__ method is a special case where we can ignore the
    #       Liskov substitution property. This is addressed in
    #       python/mypy#9254:
    #           https://github.com/python/mypy/issues/9254
    def __post_init__(self, weight: int):  # type: ignore # pylint: disable=arguments-differ
        super().__post_init__()
        self._weight = weight
        self.weight = weight # type: ignore

    def __repr__(self) -> str:
        return self.__str__()

    def desugar_weight(self, replacements: dict):
        levels = [Level(self.name) for i in range(self.weight)]
        # Multiple levels with the same name only create trouble for constraints
        # that refer to levels. When we desugar, the generated levels will not
        # have been referenced by constraints (and any reference to the old level
        # is replaced to a reference to a derived level), so duplicates are ok:
        for l in levels:
            l._synthesized = True
        return levels

@dataclass(eq=False)
class DerivedLevel(Level):
    """A :class:`.Level` representing the intersection of other levels,
    potentially from different :class:`Factors <.Factor>`. These are produced
    through use of :class:`Window <.Window>`, which are
    constraints on level selection during trial sampling.

    For more information on when to use derived levels, consult :ref:`the
    Derivations section of the SweetPea guide <guide_factorial_derivations>`.

    :param name:
        The name of the level.

    :param window:
        A :class:`.Window` used in producing a cross product of this
        level.
    """

    #: The derivation window corresponding to this level.
    window: Window = field(compare=False)

    weight: InitVar[int] = 1

    # NOTE: The __post_init__ method is a special case where we can ignore the
    #       Liskov substitution property. This is addressed in
    #       python/mypy#9254:
    #           https://github.com/python/mypy/issues/9254
    def __post_init__(self, weight: int):  # type: ignore # pylint: disable=arguments-differ
        super().__post_init__()
        # NOTE: This check is for backwards compatibility. It should instead be
        #       handled by type-checking.
        if not isinstance(self.window, Window):
            raise TypeError(f"DerivedLevel must be given a Window; got {type(self.window).__name__}.")
        self._weight = weight
        self.weight = weight # type: ignore
        # Verify internal factors' strides.
        for factor in self.window.factors:
            if (isinstance(factor, DerivedFactor)
                and factor.has_complex_window
                and factor.levels[0].window.stride > 1):
                raise ValueError(f"{type(self).__name__} does not accept factors with stride > 1, but factor "
                                 f"{factor.name} has window with stride {factor.first_level.window.stride}.")
        # Depth helps order of filling in levels when derived factors depend
        # on other derived factors
        self._depth = max(map(lambda f: f._get_depth(), self.window.factors))

    def get_dependent_cross_product(self) -> List[Tuple[Level, ...]]:
        """Produces a list of n-tuples, where each tuple represents a unique
        combination of :class:`.Level` selections among all possibilities.
        If a the derived level's start implies that not enough values are
        ready, then a ``BeforeStart`` is included in the possibilities.

        For instance, if we have two :class:`Factors <.Factor>`, each with some
        :class:`Levels <.Level>`:

        ======  ================
        Factor  Levels
        ======  ================
        color   red, blue, green
        value   1, 2
        ======  ================

        Then the following list of tuples is returned::

          [(red, 1),
           (red, 2),
           (blue, 1),
           (blue, 2),
           (green, 1),
           (green, 2)]

        .. tip::

            You can access a :class:`.Level`'s corresponding :class:`.Factor`
            via the :attr:`.Level.factor` attribute.

        :rtype: typing.List[typing.Tuple[.Level, ...]]
        """
        def levels_of(factor, i):
            ready_at = 0
            if isinstance(factor, DerivedFactor) and factor.has_complex_window:
                ready_at = factor.first_level.window.start
            else:
                ready_at = 0
            if ready_at > self.window.start - self.window.width + i + 1:
                return factor.levels + [BeforeStart(ready_at)]
            else:
                return factor.levels
        return list(product(*(levels_of(factor, i) for factor in self.window.factors for i in range(self.window.width))))

    def uses_factor(self, f: Factor):
        return any(list(map(lambda wf: wf.uses_factor(f), self.window.factors)))

    def _trial_arguments(self, sample: dict, i: int) -> list:
        """Returns the arguments used from sample trial i (zero-based) used in the level's predicate."""
        window = self.window
        args = []
        for f in window.factors:
            levels = sample[f]
            for j in range(window.width):
                idx = i-(window.width-1)+j
                if idx >= 0:
                    args.append(levels[idx].name)
                else:
                    args.append(None)
        if window.width > 1:
            args = list(chunk_dict(args, window.width))
        return args



    def __repr__(self) -> str:
        return "Derived" + self.__str__()

    def desugar_for_weights(self, replacements: dict):
        l = DerivedLevel(self.name,
                         Window(self.window.predicate,
                                [replacements.get(f, [f, f])[0] for f in self.window.factors],
                                self.window.width,
                                self.window.stride,
                                self.window.start))
        replacements[self] = l


        

@dataclass(eq=False)
class ElseLevel(Level):
    # TODO: I'm honestly not sure what this kind of level is for, semantically.
    """A :class:`.Level` for... something.

    :param name:
        The name of the level.
    """

    weight: InitVar[int] = 1

    # NOTE: The __post_init__ method is a special case where we can ignore the
    #       Liskov substitution property. This is addressed in
    #       python/mypy#9254:
    #           https://github.com/python/mypy/issues/9254
    def __post_init__(self, weight: int):  # type: ignore # pylint: disable=arguments-differ
        super().__post_init__()
        self._weight = weight
        self.weight = weight # type: ignore

    def derive_level_from_levels(self, other_levels: List[DerivedLevel]) -> DerivedLevel:
        """Converts the :class:`.ElseLevel` into a :class:`.DerivedLevel` by
        combining it with other specified
        :class:`DerivedLevels <.DerivedLevel>`.

        :param other_levels:
            A list of :class:`DerivedLevels <.DerivedLevel>` for use in
            constructing the new level.
        """
        if not other_levels:
            return DerivedLevel(self.name, WithinTrial(lambda: False, []))
        first_level = other_levels[0]
        window = Window(lambda *args: not any(map(lambda l: l.window.predicate(*args), other_levels)),
                        first_level.window.factors,
                        first_level.window.width,
                        first_level.window.stride,
                        first_level.window.start)
        return DerivedLevel(self.name, window, self.weight)

###############################################################################
##
## Factors
##

class HiddenName:
    def __init__(self, name: str):
        self.name = name

@dataclass
class Factor:
    """An independent variable in a factorial experiment. Factors are composed
    of :class:`Levels <.Level>` and come in two flavors:

    - :class:`.SimpleFactor`
    - :class:`.DerivedFactor`

    However, both classes can be implicitly constructed by the base
    :class:`.Factor` constructor, so we recommend you always use that for
    creating factors.

    During :class:`.Factor` construction, the first :class:`.Level` in the
    :attr:`~.Factor.initial_levels` is dynamically type-checked. If it's a
    :class:`.DerivedLevel` or :class:`.ElseLevel`, a :class:`.DerivedFactor` will be
    initialized. Otherwise, a :class:`.SimpleFactor` will be initialized.

    In all cases, the :attr:`~.Factor.initial_levels` will be processed. This
    step ensures that all of a factor's :attr:`~.Factor.levels` will always be
    :class:`Levels <.Level>`. The levels are processed according to these
    rules:

    1. A :class:`.Level` instance is left alone.

    2. A :class:`str` instance is converted into a :class:`.SimpleLevel`.

    3. A :class:`tuple` or :class:`list` consisting of exactly one :class:`str`
       followed by one :class:`int` will be converted into a
       :class:`.SimpleLevel` with the :class:`str` as its name and the
       :class:`int` as its weight.

    4. Anything else is converted into a :class:`.SimpleLevel` by using it
       as a level name.

    .. note::

        The :class:`.DerivedFactor` subclass does additional processing after
        these steps.

    :param name:
        The name of this factor.

    :param initial_levels:
        The levels comprising this factor. The list can be made of anything,
        but any values in the list that are not instances of :class:`.Level` or
        one of its subclasses will be converted into :class:`.SimpleLevel`
        instances using ``SimpleLevel(value)``.
    :type initial_levels: typing.Sequence[Any]

    :rtype: .Factor

    .. tip::

        See :ref:`the Factorial Experiment Design section of the SweetPea guide
        <guide_factorial_design>` for more about factors, levels, and how to
        use them.
    """

    #: The name of this factor. A synthesized factor may be mutated to hide its name
    name: Union[str, HiddenName]

    #: The levels used during factor initialization.
    initial_levels: InitVar[Sequence[Any]]

    #: The discrete values that this factor can have.
    levels: Sequence[Level] = field(init=False)

    #: A mapping from level names to levels for constant-time lookup.
    _level_map: Dict[str, Level] = field(init=False, default_factory=dict)

    def __new__(cls, name: Union[str, HiddenName], initial_levels: Sequence[Any], *_, **__) -> Factor:
        # Ensure we got a string for a name. This requirement is imposed for
        # backwards compatibility, but it should be handled by type-checking.
        if not isinstance(name, (str, HiddenName)):
            raise ValueError(f"Factor name not a string: {name}.")
        # Check if we're initializing this from `Factor` directly or one of its
        # subclasses.
        if cls != Factor:
            # It's a subclass, so we do nothing special.
            return super().__new__(cls)
        # Otherwise, we have to check whether to return a subclass instance.
        # This requires there to be at least 1 initial level.
        if not initial_levels:
            raise ValueError(f"Expected at least one level for factor {name}.")
        first_level = initial_levels[0]
        if isinstance(first_level, (DerivedLevel, ElseLevel)):
            instance = super().__new__(DerivedFactor)
        else:
            instance = super().__new__(SimpleFactor)
        return instance

    def __post_init__(self, initial_levels: Sequence[Any]):
        # First, we convert the given initial levels into actual `Level`s.
        real_levels: List[Level] = []
        for level in initial_levels:
            if isinstance(level, Level):
                pass
            else:
                level = SimpleLevel(level)
            real_levels.append(level)
        # Then we do any necessary post-processing of the levels.
        self.levels = self._process_initial_levels(real_levels)
        for level in self.levels:
            if hasattr(level, "factor"):
                raise ValueError(f"Level already belongs to a factor: {level.name}")
            if level.name in self._level_map and not level._synthesized:
                raise ValueError(f"Multiple levels with the same name: {level.name}")
            self._level_map[level.name] = level
            level.factor = self

    # NOTE: Subclasses of `Factor` must override this method!
    # NOTE: This method cannot be decorated `@abstractmethod` because the
    #       `abc.ABC` class does not play well with `@dataclass`. Additionally,
    #       `Factor` is not actually an abstract base class because the custom
    #       `__new__` implementation prevents it from ever being instantiated.
    #       There is no way to express "a class which cannot be instantiated"
    #       that I know of, so there's no way to get around this dynamic
    #       `NotImplementedError` solution.
    def _process_initial_levels(self, initial_levels: Sequence[Level]) -> Sequence[Level]:
        raise NotImplementedError

    def __deepcopy__(self, memo: Dict):
        cls = self.__class__
        new_instance = cls.__new__(cls, self.name, [])
        memo[id(self)] = new_instance
        for attr, val in self.__dict__.items():
            setattr(new_instance, attr, deepcopy(val, memo))
        return new_instance

    def __eq__(self, other) -> bool:
        # only equal when it's the same object:
        return id(self) == id(other)

    def __str__(self) -> str:
        # levels_string = '[' + ', '.join(map(str, self.levels)) + ']'
        # return f"Factor<{self.name} | {levels_string}>"
        return f"Factor<{self.name}>"

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.name)

    def __getitem__(self, name) -> Level:
        value = self.get_level(name)
        if value is None:
            raise KeyError(f"Factor {self.name} has no level named {name}.")
        return value

    def __contains__(self, name) -> bool:
        return name in self.levels

    def get_level(self, name) -> Optional[Level]:
        """Returns a :class:`.Level` instance corresponding to the given name,
        if it exists within this factor. Otherwise, returns ``None``.
        """
        return self._level_map.get(name)

    # TODO: This should be made private.
    @property
    def first_level(self) -> Level:
        """The first :class:`.Level` in this factor."""
        return self.levels[0]

    @property
    def has_complex_window(self) -> bool:
        """Whether this factor has a complex derivation window.

        A complex derivation window is a window  of derived factors whose
        first-level derivations are themselves considered complex.
        """
        if not isinstance(self, DerivedFactor):
            return False
        window = self.first_level.window
        return (window.width > 1
                or window.stride > 1
                or cast(int, window.start) > 0
                or window.is_complex)

    def applies_to_trial(self, trial_number: int) -> bool:
        """Whether this factor applies to the given trial. For example, factors
        with :class:`.TransitionDerivation` derivations in their levels do not
        apply to trial number ``1``, but do apply to all subsequent trials.

        .. tip::

            Trials start their numbering at ``1``.
        """
        if trial_number <= 0:
            raise ValueError(f"Trial numbers must be 1 or greater; got {trial_number}.")
        if not isinstance(self, DerivedFactor):
            return True

        window = self.first_level.window
        return (trial_number >= (cast(int, window.start)+1)
                and (trial_number - (cast(int, window.start)+1)) % window.stride == 0)


    def uses_factor(self, f: Factor):
        return self == f

    def _get_depth(self):
        if not isinstance(self, DerivedFactor):
            return 0
        return max(map(lambda l: cast(DerivedLevel, l)._depth, self.levels))+1

    def level_weight_sum(self):
        return sum([l.weight for l in self.levels])

    def test_trial(self, i, trial_sequence):
        """Test if trial i of trial_sequence meets the criteria
        If this is a regular factor this is always True
        """
        return True

@dataclass
class SimpleFactor(Factor):
    """A :class:`.Factor` comprised of :class:`SimpleLevels <.SimpleLevel>`. If
    a subclass of :class:`.Level` is passed in that is not a
    :class:`SimpleLevel`, an error is raised during initialization.

    :param name:
        The name of this factor.

    :param initial_levels:
        The levels comprising this factor.
    :type initial_levels: typing.Sequence[.SimpleLevel]

    :rtype: .SimpleFactor
    """

    levels: Sequence[SimpleLevel] = field(init=False)

    def _process_initial_levels(self, initial_levels) -> Sequence[SimpleLevel]:
        for level in initial_levels:
            if not isinstance(level, SimpleLevel):
                raise ValueError(f"{type(self).__name__} named {self.name} must contain only SimpleLevels, but level "
                                 f"{level.name} has type {type(level).__name__}.")
        return initial_levels

    # NOTE: Pylint incorrectly labels this as a "useless super delegation"
    #       because it is unaware that dataclasses do not automatically inherit
    #       __hash__, __eq__, etc in certain cases. This is addressed in
    #       PyCQA/pylint#3934:
    #           https://github.com/PyCQA/pylint/issues/3934
    def __hash__(self) -> int:  # pylint: disable=useless-super-delegation
        return super().__hash__()

    # NOTE: See note on `SimpleFactor.__hash__`.
    def __eq__(self, other) -> bool:  # pylint: disable=useless-super-delegation
        return super().__eq__(other)

    def __repr__(self) -> str:
        return self.__str__()

    def get_level(self, name) -> Optional[SimpleLevel]:
        return cast(Optional[SimpleLevel], super().get_level(name))

    @property
    def first_level(self) -> SimpleLevel:
        return cast(SimpleLevel, self.levels[0])

    def desugar_weights(self, replacements: dict):
        levelss = [l.desugar_weight(replacements) for l in self.levels]
        flat_f = Factor(self.name, list(chain.from_iterable(levelss)))
        names = list(map(lambda name: (name, DerivedLevel(name, WithinTrial(lambda n: n == name, [flat_f]))),
                         [l.name for l in self.levels]))
        derived_levels = {name: derived_level for name, derived_level in names}
        for l in self.levels:
            replacements[l] = derived_levels[l.name]
        derived_f = DerivedFactor(HiddenName(cast(str, self.name)), list(derived_levels.values()))
        replacements[self] = [derived_f, flat_f]

@dataclass
class DerivedFactor(Factor):
    """A :class:`.Factor` composed of :class:`DerivedLevels <.DerivedLevel>`.

    After doing the level processing described by :class:`.Factor`, an
    additional step is taken: all :class:`ElseLevels <.ElseLevel>` are
    converted into :class:`DerivedLevels <.DerivedLevel>`. This is done by
    calling :func:`.ElseLevel.derive_level_from_levels`, supplying all the
    natural :class:`DerivedLevels <.DerivedLevel>` that were passed in.

    :param name:
        The name of this factor.

    :param initial_levels:
        The levels comprising this factor. The levels may be instances of
        either :class:`DerivedLevels <.DerivedLevel>` or
        :class:`ElseLevels <.ElseLevel>`.

        .. note::

            Any given :class:`.ElseLevel` will be converted to a
            :class:`.DerivedLevel` by way of
            :func:`.ElseLevel.derive_level_from_levels`, so that the
            :attr:`.Factor.levels` list will only contain
            :class:`DerivedLevels <DerivedLevel>`.
    :type initial_levels: typing.Sequence[typing.Union[.DerivedLevel, .ElseLevel]]

    :rtype: .DerivedFactor
    """

    levels: Sequence[DerivedLevel] = field(init=False)

    def _process_initial_levels(self, initial_levels: Sequence[Level]) -> Sequence[DerivedLevel]:
        # First, we construct the list of `DerivedLevel`s.
        adjusted_levels: List[DerivedLevel] = []
        initial_derived_levels: List[DerivedLevel] = [level for level in initial_levels
                                                      if isinstance(level, DerivedLevel)]
        for level in initial_levels:
            if isinstance(level, ElseLevel):
                adjusted_levels.append(level.derive_level_from_levels(initial_derived_levels))
            elif isinstance(level, DerivedLevel):
                adjusted_levels.append(level)
            else:
                raise ValueError(f"Cannot use {type(level).__name__} to create a {type(self).__name__}. "
                                 f"Only DerivedLevels and ElseLevels are allowed.")
        # Then, we do some validation.
        expected_size = adjusted_levels[0].window.size
        for level in adjusted_levels[1:]:
            if level.window.size != expected_size:
                raise ValueError(f"Expected all DerivedLevels' derivations in factor {self.name} to have size "
                                 f"{expected_size}, but level {level.name} has a derivation window of size "
                                 f"{level.window.size}.")
        expected_factors = adjusted_levels[0].window.factors
        for level in adjusted_levels[1:]:
            if level.window.factors != expected_factors:
                raise ValueError(f"Expected all DerivedLevels' derivations in factor {self.name} to use factors "
                                 f"{expected_factors}, but level {level.name}'s derivation window uses factors "
                                 f"{level.window.factors}.")
        return adjusted_levels

    # NOTE: See note on `SimpleFactor.__hash__`.
    def __hash__(self) -> int:  # pylint: disable=useless-super-delegation
        return super().__hash__()

    # NOTE: See note on `SimpleFactor.__hash__`.
    def __eq__(self, other) -> bool:  # pylint: disable=useless-super-delegation
        return super().__eq__(other)

    def __repr__(self) -> str:
        return self.__str__()

    def get_level(self, name) -> Optional[DerivedLevel]:
        return cast(Optional[DerivedLevel], super().get_level(name))

    @property
    def first_level(self) -> DerivedLevel:
        return cast(DerivedLevel, self.levels[0])

    def uses_factor(self, f: Factor):
        return (self == f) or any(list(map(lambda l: l.uses_factor(f), self.levels)))

    def select_level_for_sample(self, i: int, sample: dict) -> Any:
        """Get level name for trial i (zero-based) depending on
        values of other factors already in the sample."""
        args = self.levels[0]._trial_arguments(sample, i)
        for l in self.levels:
            if l.window.predicate(*args):
                return l
        raise RuntimeError("no matching trial found when filling in a sample")

    def desugar_for_weights(self, replacements: dict):
        if any([any([f in replacements for f in l.window.factors]) for l in self.levels]):
            for l in self.levels:
                l.desugar_for_weights(replacements)
            f = DerivedFactor(self.name, [replacements[l] for l in self.levels])
            replacements[self] = [f, f]

    def test_trial(self, i, trial_sequence):
        res = True
        for level in self.levels:
            if trial_sequence[self][i] == level:
                args = level._trial_arguments(trial_sequence, i)
                res &= level.window.predicate(*args)
        return res





###############################################################################
##
## Derivation Windows
##


@dataclass
class Window:
    """A mechanism to specify the window of a derivation. Derivation windows
    are used to inform a derivation based on levels from other factors in the
    current trial and, potentially, multiple preceding trials.

    For detailed information, see the SweetPea guide's :ref:`section on
    derivation <guide_factorial_derivations>`.

    :param predicate:
        A predicate function used during derivation. Different derivation
        windows require different forms of predicates.

        .. warning::

            The type of arguments passed to the ``predicate`` varies by
            Derivation subclass. Read their documentation carefully!
    :type predicate: typing.Callable[[Any, ....], bool]

    :param factors:
        The factors upon which this derivation window depends.
    :type factors: typing.Sequence[.Factor]

    :param width:
        The number of trials that this derivation window depends upon. The
        current trial must be included, so the number must be greater than or
        equal to ``1``.
    :type width: int

    :param stride:
        How often the factor is derived. With a stride of ``1`` (the default),
        the factor is derived for every trial starting with initial one. For
        a larger stride, ``stride-1`` trials are skipped before the factor is
        derived again.
    :type stride: int

    :param start:
        The first trial where this derivation applies. By default, the starting
        trial is the first trial where the factors are all defined that this one
        depends on. If the starting trial is before that, then the predicate
        must be prepared to handle ``None`` as an argument level.
    :type start: int
    """

    # TODO: I think that, if possible, this should be changed to some other
    #       managed data structure (like an enum that correlates cases to
    #       predicates). That seems like it would be more robust and also
    #       provide an easier API for users of SweetPea.
    # TODO: Alternatively, we could just make this class-private and provide a
    #       method that takes the arguments and returns the result.
    #: The predicate used during derivation with this window.
    predicate: Callable

    #: The factors upon which this derivation depends.
    factors: List[Factor]

    # TODO: Rename this to something more clear.
    #: The width of this window.
    width: int

    #: The stride of this window.
    stride: int = 1

    # The starting trial, if not automatic; 0-based
    start: Optional[int] = None

    def __post_init__(self):
        # NOTE: We check the types for backwards compatibility, but it should
        #       be handled with type-checking.
        if not callable(self.predicate):
            raise TypeError(f"Window expected predicate function; got {self.predicate}.")
        
        for factor in self.factors:
            if not isinstance(factor, Factor):
                raise TypeError(f"Window must be composed of Factors; got {factor}.")
        # TODO: Validate the predicate accepts the same number of arguments as
        #       factors in `self.factors`.
        if self.width < 1:
            raise ValueError(f"A {type(self).__name__} derivation window must have a width of at least 1.")
        if self.stride < 1:
            raise ValueError(f"A {type(self).__name__} derivation window must have a stride of at least 1.")

        default_start = self.width - 1
        for factor in self.factors:
            ready_at = 0
            if isinstance(factor, DerivedFactor) and factor.has_complex_window:
                ready_at = factor.first_level.window.start
            ready_at = ready_at + self.width - 1
            if ready_at > default_start:
                default_start = ready_at
        if (self.start == None):
            self.start = default_start
        elif self.start < 0:
            raise ValueError(f"A {type(self).__name__} derivation window must have a start of at least 0.")
        self.start_delta = self.start - default_start
        found_factor_names = set()
        for factor in self.factors:
            if factor.name in found_factor_names:
                raise ValueError(f"Derivations do not accept repeated factors. Factor repeated: {factor.name}.")
            found_factor_names.add(factor.name)

    @property
    def size(self) -> Tuple[int, int, int]:
        """The width and stride of this derivation window as a pair.

        :rtype: typing.Tuple[int, int, int]
        """
        return (self.width, self.stride, cast(int, self.start))

    @property
    def first_factor(self) -> Factor:
        """The first :class:`.Factor` in the internal list of Factors.

        :rtype: .Factor
        """
        return self.factors[0]

    @property
    def is_complex(self) -> bool:
        """Whether this derivation window is considered *complex*. A complex
        derivation windowis one for which at least one of the following is
        true:

        - The window's :attr:`.width` is greater than ``1``.
        - The window's :attr:`.stride` is greater than ``1``.
        - The window's :attr:`.first_factor` is a complex derived factor.

        :rtype: bool
        """
        return (self.width > 1
                or self.stride > 1
                or self.first_factor.has_complex_window)

@dataclass
class WithinTrial(Window):
    """A derivation window that depends on levels from other factors, all
    within the same trial.

    :param predicate:
        A predicate that takes as many :class:`strs <str>` as factors in this
        derivation window, where each :class:`str` will be the name of a
        factor.
    :type predicate: typing.Callable[[str, ....], bool]

    :param factors:
        The factors upon which this derivation window depends.
    :type factors: typing.Sequence[.Factor]
    """

    width: int = field(default=1, init=False)
    stride: int = field(default=1, init=False)
    start: Optional[int] = field(default=None, init=False)


@dataclass
class Transition(Window):
    """A derivation window that depends on levels from other factors in the
    current trial and the immediately preceding trial.
    :class:`TransitionDerivations <.TransitionDerivation>` are used to
    constrain the transition points between trials.

    :param predicate:
        A predicate that takes as many lists of factors as factors in this
        derivation window.
    :type predicate: typing.Callable[[typing.Sequence[.Factor], ....], bool]

    :param factors:
        The factors upon which this derivation window depends.
    :type factors: typing.Sequence[.Factor]

    """

    width: int = field(default=2, init=False)
    stride: int = field(default=1, init=False)
    start: int = field(default=1, init=False)



import operator as op
import pytest

from sweetpea import *

color = Factor("color", ['red', 'green'])
size = Factor("size", ['small', 'med', 'large'])

red = color.get_level('red')
med = size.get_level('med')

match2 = Factor(name="match2", initial_levels=[
    DerivedLevel("up2", Window(lambda a: a[-2] == 'red', [color], 3, 1)),
    DerivedLevel("down2", Window(lambda a: a[-2] != 'red', [color], 3, 1))
])

match1 = Factor(name="match1", initial_levels=[
    DerivedLevel("up2", Window(lambda a: a[-1] == 'red', [color], 2, 1)),
    DerivedLevel("down2", Window(lambda a: a[-1] != 'red', [color], 2, 1))
])

def check_crossed_derived_factor(strategy, design, solutions, constraints=[], repeat_constraints=[]):
    block = CrossBlock(design=design,
                       crossing=design,
                       constraints=constraints)
    repeat = Repeat(block, repeat_constraints)

    experiments = synthesize_trials(block=repeat, samples=1000, sampling_strategy=strategy)
    
    assert len(experiments) == solutions

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_check_small_crossed_derived_factor(strategy):
    check_crossed_derived_factor(strategy, [color, match2], 4)

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_check_big_crossed_derived_factor(strategy):
    check_crossed_derived_factor(strategy, [color, match2, match1], 16)

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_check_crossed_derived_factor_with_minimum_trials(strategy):
    check_crossed_derived_factor(strategy, [color, match2], 108, [MinimumTrials(8)])
    check_crossed_derived_factor(strategy, [color, match2], 16, [], [MinimumTrials(8)])



"""This module provides a simple enumeration base class for managing return
codes in command-line utilities.
"""


from enum import Enum


__all__ = ['ReturnCodeEnum']


class ReturnCodeEnum(Enum):
    """Command-line functions produce return codes, and sometimes these do not
    necessarily indicate an error. Subclasses of this enum can be used for
    managing the various return codes and checking against them conveniently.
    """

    @classmethod
    def has_value(cls, value: int) -> bool:
        """Determines whether this :class:`ReturnCodeEnum` contains the
        indicated ``value``.
        """
        return value in (e.value for e in cls)



"""This module provides uniform CNF sampling functionality through the
:func:`sample_uniform` function.
"""


from typing import List

from ..cnf import CNF
from .tools.unigen import DEFAULT_DOCKER_MODE_ON, call_unigen
from .utility import GenerationRequest, Solution, combine_and_save_cnf, temporary_cnf_file


__all__ = ['sample_uniform']


def sample_uniform(sample_count: int,
                    initial_cnf: CNF,
                   fresh: int,
                   support: int,
                   generation_requests: List[GenerationRequest],
                   use_docker: bool = DEFAULT_DOCKER_MODE_ON,
                   use_cmsgen: bool = False
                   ) -> List[Solution]:
    """Samples solutions to a CNF problem uniformly. The solution is computed
    using Unigen.
    """
    with temporary_cnf_file() as cnf_file:
        combine_and_save_cnf(cnf_file, initial_cnf, fresh, support, generation_requests)
        solver_name = "UniGen" if not use_cmsgen else "CMSGen"
        print(f"Running {solver_name}...")
        solution_str = call_unigen(sample_count, cnf_file, docker_mode=use_docker, use_cmsgen=use_cmsgen)
        # TODO: Validate that skipping the comments is the intended
        #       functionality. The Haskell code doesn't appear to need to do
        #       this, but this could be due to the Unigen upgrade or something
        #       else. Just check it.
        if not solution_str:
            return []
        sample_set = 0
        if "we found only " in solution_str:
            sample_set = int(solution_str[solution_str.index("we found only ")+14:].split(',')[0])

        return [build_solution(line) for line in solution_str.strip().splitlines() if line and not line.startswith('c')][sample_set:]


def build_solution(line: str) -> Solution:
    """Given a Unigen solution, constructs a :class:`.Solution` object."""
    parts = line.replace('v', '').strip().split()
    assignment = [int(p) for p in parts[:-1]]
    frequency = int(parts[-1].split(':')[-1])
    return Solution(assignment, frequency)



from typing import List
from sweetpea._internal.primitive import Factor, Level


def _convert_from_name_to_factor(name: str, design: List[Factor]) -> Factor:
    """
    given a design, convert a string that indicates the name of a factor to the matching Factor in the design
    """
    # get factor names from design
    for f in design:
        if name == f.name:
            return f
    raise Exception(f'Error while converting {name}: The given factor name is not in the list of Factors')


def _convert_form_name_to_level(name: str, factor: Factor) -> Level:
    """
    given a factor, convert a string that indicates the name of a level to the matching level of the Factor
    """
    if not name:
        return Level('')
    for l in factor.levels:
        if name == l.name:
            return l
    raise Exception(f'Error while converting {name}: The given level name is not a Factor level')


def convert_sample_from_names_to_objects(sample: dict, design: List[Factor]) -> dict:
    """
    given a design, converts a sample formatted in factor names and level names into factor objects and level objects
    """
    new_dict = {}
    for factor_name in sample.keys():
        factor = _convert_from_name_to_factor(factor_name, design)
        value = [_convert_form_name_to_level(level_name, factor) for level_name in sample[factor_name]]
        new_dict[factor] = value
    return new_dict



"""This module provides aliases for values used in representing numbers in a
binary format.
"""


from typing import List


__all__ = ['BinaryNumber', 'int_to_binary']


#: A representation of a binary number as a list of integers, where each
#: element of the list is either 1 or -1.
BinaryNumber = List[int]


def int_to_binary(value: int) -> BinaryNumber:
    """Converts an integer value to binary where the output representation is a
    list of binary digits of the alphabet {-1, 1} (i.e., similar to a normal
    binary representation except the "false" value is -1 instead of 0).

    For example:

    =======  =================  ==========
    Decimal  ``int_to_binary``  Binary
    =======  =================  ==========
    ``2``    ``[1, -1]``        ``0b10``
    ``11``   ``[1, -1, 1, 1]``  ``0b1011``
    =======  =================  ==========
    """
    output: BinaryNumber = []
    while value != 0:
        if value % 2 == 0:
            output.append(-1)
        else:
            output.append(1)
        value //= 2
    output.reverse()
    return output



import operator as op
import pytest

from sweetpea import CrossBlock
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial
from sweetpea._internal.constraint import Reify
from sweetpea._internal.sampling_strategy.random import UCSolutionEnumerator

color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])
congruency = Factor("congruency", [
    DerivedLevel("congruent",   WithinTrial(op.eq, [color, text])),
    DerivedLevel("incongruent", WithinTrial(op.ne, [color, text]))
])

crossing = [color, text]
design = [color, text]
block = CrossBlock(design, crossing, [])


def test_generate_crossing_instances():
    enumerator = UCSolutionEnumerator(block)
    crossing_instances = enumerator._UCSolutionEnumerator__generate_crossing_instances()
    simplified_names = []
    for d in crossing_instances:
        d_simple = {}
        for (f, l) in d.items():
            d_simple[f.name] = l.name
        simplified_names.append(d_simple)

    assert simplified_names == [
        {'color': 'red', 'text': 'red'},
        {'color': 'red', 'text': 'blue'},
        {'color': 'blue', 'text': 'red'},
        {'color': 'blue', 'text': 'blue'}
    ]


def test_generate_source_combinations():
    block = CrossBlock([color, text, congruency], [congruency], list(map(Reify, design)))
    enumerator = UCSolutionEnumerator(block)
    crossing_source_combos = enumerator._UCSolutionEnumerator__generate_source_combinations()
    simplified_names = []
    for d in crossing_source_combos:
        d_simple = {}
        for (f, l) in d.items():
            d_simple[f.name] = l.name
        simplified_names.append(d_simple)

    assert simplified_names == [
        {'color': 'red', 'text': 'red'},
        {'color': 'red', 'text': 'blue'},
        {'color': 'blue', 'text': 'red'},
        {'color': 'blue', 'text': 'blue'}
    ]


@pytest.mark.parametrize('sequence_number, expected_solution', [
[0, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['red', 'blue', 'red', 'blue']}],
[1, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['blue', 'red', 'red', 'blue']}],
[2, {'color': ['blue', 'red', 'red', 'blue'], 'text': ['red', 'red', 'blue', 'blue']}],
[3, {'color': ['blue', 'red', 'red', 'blue'], 'text': ['blue', 'red', 'blue', 'red']}],
[4, {'color': ['red', 'blue', 'red', 'blue'], 'text': ['red', 'red', 'blue', 'blue']}],
[5, {'color': ['red', 'blue', 'red', 'blue'], 'text': ['blue', 'red', 'red', 'blue']}],
[6, {'color': ['blue', 'red', 'red', 'blue'], 'text': ['red', 'blue', 'red', 'blue']}],
[7, {'color': ['blue', 'red', 'red', 'blue'], 'text': ['blue', 'blue', 'red', 'red']}],
[8, {'color': ['red', 'blue', 'red', 'blue'], 'text': ['red', 'blue', 'blue', 'red']}],
[9, {'color': ['red', 'blue', 'red', 'blue'], 'text': ['blue', 'blue', 'red', 'red']}],
[10, {'color': ['blue', 'blue', 'red', 'red'], 'text': ['red', 'blue', 'red', 'blue']}],
[11, {'color': ['blue', 'blue', 'red', 'red'], 'text': ['blue', 'red', 'red', 'blue']}],
[12, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['red', 'blue', 'blue', 'red']}],
[13, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['blue', 'red', 'blue', 'red']}],
[14, {'color': ['blue', 'red', 'blue', 'red'], 'text': ['red', 'red', 'blue', 'blue']}],
[15, {'color': ['blue', 'red', 'blue', 'red'], 'text': ['blue', 'red', 'red', 'blue']}],
[16, {'color': ['red', 'blue', 'blue', 'red'], 'text': ['red', 'red', 'blue', 'blue']}],
[17, {'color': ['red', 'blue', 'blue', 'red'], 'text': ['blue', 'red', 'blue', 'red']}],
[18, {'color': ['blue', 'red', 'blue', 'red'], 'text': ['red', 'blue', 'blue', 'red']}],
[19, {'color': ['blue', 'red', 'blue', 'red'], 'text': ['blue', 'blue', 'red', 'red']}],
[20, {'color': ['red', 'blue', 'blue', 'red'], 'text': ['red', 'blue', 'red', 'blue']}],
[21, {'color': ['red', 'blue', 'blue', 'red'], 'text': ['blue', 'blue', 'red', 'red']}],
[22, {'color': ['blue', 'blue', 'red', 'red'], 'text': ['red', 'blue', 'blue', 'red']}],
[23, {'color': ['blue', 'blue', 'red', 'red'], 'text': ['blue', 'red', 'blue', 'red']}]
])
def test_generate_sample_basic_stroop(sequence_number, expected_solution):
    block = CrossBlock([color, text, congruency],
                       [color, text],
                       [Reify(congruency)]) # even so, will not show up in enumerator output
    enumerator = UCSolutionEnumerator(block)
    assert enumerator.factors_and_levels_to_names(enumerator.generate_sample(sequence_number)) == expected_solution

@pytest.mark.parametrize('sequence_number, expected_solution', [
[0, {'color': ['red', 'red', 'blue', 'blue'], 'congruency': ['congruent', 'incongruent', 'congruent', 'incongruent'], 'text': ['red', 'blue', 'blue', 'red']}],
[1, {'color': ['red', 'red', 'blue', 'blue'], 'congruency': ['incongruent', 'congruent', 'congruent', 'incongruent'], 'text': ['blue', 'red', 'blue', 'red']}],
[2, {'color': ['blue', 'red', 'red', 'blue'], 'congruency': ['congruent', 'congruent', 'incongruent', 'incongruent'], 'text': ['blue', 'red', 'blue', 'red']}],
[3, {'color': ['blue', 'red', 'red', 'blue'], 'congruency': ['incongruent', 'congruent', 'incongruent', 'congruent'], 'text': ['red', 'red', 'blue', 'blue']}],
[4, {'color': ['red', 'blue', 'red', 'blue'], 'congruency': ['congruent', 'congruent', 'incongruent', 'incongruent'], 'text': ['red', 'blue', 'blue', 'red']}],
[5, {'color': ['red', 'blue', 'red', 'blue'], 'congruency': ['incongruent', 'congruent', 'congruent', 'incongruent'], 'text': ['blue', 'blue', 'red', 'red']}],
[6, {'color': ['blue', 'red', 'red', 'blue'], 'congruency': ['congruent', 'incongruent', 'congruent', 'incongruent'], 'text': ['blue', 'blue', 'red', 'red']}],
[7, {'color': ['blue', 'red', 'red', 'blue'], 'congruency': ['incongruent', 'incongruent', 'congruent', 'congruent'], 'text': ['red', 'blue', 'red', 'blue']}],
[8, {'color': ['red', 'blue', 'red', 'blue'], 'congruency': ['congruent', 'incongruent', 'incongruent', 'congruent'], 'text': ['red', 'red', 'blue', 'blue']}],
[9, {'color': ['red', 'blue', 'red', 'blue'], 'congruency': ['incongruent', 'incongruent', 'congruent', 'congruent'], 'text': ['blue', 'red', 'red', 'blue']}],
[10, {'color': ['blue', 'blue', 'red', 'red'], 'congruency': ['congruent', 'incongruent', 'congruent', 'incongruent'], 'text': ['blue', 'red', 'red', 'blue']}],
[11, {'color': ['blue', 'blue', 'red', 'red'], 'congruency': ['incongruent', 'congruent', 'congruent', 'incongruent'], 'text': ['red', 'blue', 'red', 'blue']}],
[12, {'color': ['red', 'red', 'blue', 'blue'], 'congruency': ['congruent', 'incongruent', 'incongruent', 'congruent'], 'text': ['red', 'blue', 'red', 'blue']}],
[13, {'color': ['red', 'red', 'blue', 'blue'], 'congruency': ['incongruent', 'congruent', 'incongruent', 'congruent'], 'text': ['blue', 'red', 'red', 'blue']}],
[14, {'color': ['blue', 'red', 'blue', 'red'], 'congruency': ['congruent', 'congruent', 'incongruent', 'incongruent'], 'text': ['blue', 'red', 'red', 'blue']}],
[15, {'color': ['blue', 'red', 'blue', 'red'], 'congruency': ['incongruent', 'congruent', 'congruent', 'incongruent'], 'text': ['red', 'red', 'blue', 'blue']}],
[16, {'color': ['red', 'blue', 'blue', 'red'], 'congruency': ['congruent', 'congruent', 'incongruent', 'incongruent'], 'text': ['red', 'blue', 'red', 'blue']}],
[17, {'color': ['red', 'blue', 'blue', 'red'], 'congruency': ['incongruent', 'congruent', 'incongruent', 'congruent'], 'text': ['blue', 'blue', 'red', 'red']}],
[18, {'color': ['blue', 'red', 'blue', 'red'], 'congruency': ['congruent', 'incongruent', 'incongruent', 'congruent'], 'text': ['blue', 'blue', 'red', 'red']}],
[19, {'color': ['blue', 'red', 'blue', 'red'], 'congruency': ['incongruent', 'incongruent', 'congruent', 'congruent'], 'text': ['red', 'blue', 'blue', 'red']}],
[20, {'color': ['red', 'blue', 'blue', 'red'], 'congruency': ['congruent', 'incongruent', 'congruent', 'incongruent'], 'text': ['red', 'red', 'blue', 'blue']}],
[21, {'color': ['red', 'blue', 'blue', 'red'], 'congruency': ['incongruent', 'incongruent', 'congruent', 'congruent'], 'text': ['blue', 'red', 'blue', 'red']}],
[22, {'color': ['blue', 'blue', 'red', 'red'], 'congruency': ['congruent', 'incongruent', 'incongruent', 'congruent'], 'text': ['blue', 'red', 'blue', 'red']}],
[23, {'color': ['blue', 'blue', 'red', 'red'], 'congruency': ['incongruent', 'congruent', 'incongruent', 'congruent'], 'text': ['red', 'blue', 'blue', 'red']}]
])
def test_generate_sample_basic_stroop_variation(sequence_number, expected_solution):
    block = CrossBlock([color, text, congruency],
                       [color, congruency],
                       [])
    enumerator = UCSolutionEnumerator(block)
    assert enumerator.factors_and_levels_to_names(enumerator.generate_sample(sequence_number)) == expected_solution

# ---------------------------------------------------------------
# |   Trial |  color   |      text      |      congruency       |
# |       # | red blue | red blue green | congruent incongruent |
# ---------------------------------------------------------------
# |       1 |  1   2   |  3   4     5   |     6          7      |
# |       2 |  8   9   | 10   11   12   |    13         14      |
# |       3 | 15   16  | 17   18   19   |    20         21      |
# |       4 | 22   23  | 24   25   26   |    27         28      |
# ---------------------------------------------------------------
@pytest.mark.parametrize('sequence_number, expected_solution', [
[0, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['red', 'blue', 'blue', 'red'], 'congruency': ['congruent', 'incongruent', 'congruent', 'incongruent']}],
[1, {'color': ['red', 'red', 'blue', 'blue'], 'text': ['blue', 'red', 'blue', 'red'], 'congruency': ['incongruent', 'congruent', 'congruent', 'incongruent']}],
[41, {'color': ['red', 'blue', 'blue', 'red'], 'text': ['green', 'blue', 'red', 'red'], 'congruency': ['incongruent', 'congruent', 'incongruent', 'congruent']}]
])
def test_generate_sample_basic_stroop_uneven_colors(sequence_number, expected_solution):
    text = Factor("text", ["red", "blue", "green"])
    congruency = Factor("congruency", [
        DerivedLevel("congruent",   WithinTrial(op.eq, [color, text])),
        DerivedLevel("incongruent", WithinTrial(op.ne, [color, text]))
    ])

    block = CrossBlock([color, text, congruency],
                       [color, congruency],
                       [])
    enumerator = UCSolutionEnumerator(block)
    assert enumerator.factors_and_levels_to_names(enumerator.generate_sample(sequence_number)) == expected_solution



import operator as op
import pytest

from random import shuffle

from sweetpea._internal.sampling_strategy.base import Gen
from sweetpea import CrossBlock
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import Reify


# Common variables for stroop.
color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("color repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

text_repeats_factor = Factor("text repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [text])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [text]))
])

congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])

design = [color, text, con_factor]
crossing = [color, text]
blk = CrossBlock(design, crossing, [Reify(con_factor)])


def test_decode():
    solution = [-1,   2,  -3,   4,   5,  -6,
                -7,   8,   9, -10, -11,  12,
                13, -14, -15,  16, -17,  18,
                19, -20,  21, -22,  23, -24]
    shuffle(solution)

    assert Gen.decode(blk, solution) == {
        'color':      ['blue', 'blue', 'red',  'red'],
        'text':       ['blue', 'red',  'blue', 'red'],
        'congruent?': ['con',  'inc',  'inc',  'con']
    }

    solution = [ -1,   2, -3,   4,   5,  -6,
                  7,  -8, -9,  10, -11,  12,
                 13, -14, 15, -16,  17, -18,
                -19,  20, 21, -22, -23,  24]
    shuffle(solution)

    assert Gen.decode(blk, solution) == {
        'color':      ['blue', 'red',  'red', 'blue'],
        'text':       ['blue', 'blue', 'red', 'red'],
        'congruent?': ['con',  'inc',  'con', 'inc']
    }

    solution = [-1,   2,   3,  -4,  -5,   6,
                -7,   8,  -9,  10,  11, -12,
                13, -14,  15, -16,  17, -18,
                19, -20, -21,  22, -23,  24]
    shuffle(solution)

    assert Gen.decode(blk, solution) == {
        'color':      ['blue', 'blue', 'red', 'red'],
        'text':       ['red',  'blue', 'red', 'blue'],
        'congruent?': ['inc',  'con',  'con', 'inc']
    }

    f1 = Factor("a", ["b", "c", "d"])
    f2 = Factor("e", ["f"])
    f_blk = CrossBlock([f1, f2], [f1, f2], [])
    solution = [-1,   2,  -3, 4,
                -5,  -6,   7, 8,
                 9, -10  -11, 12]
    shuffle(solution)

    assert Gen.decode(f_blk, solution) == {
        'a': ['c', 'd', 'b'],
        'e': ['f', 'f', 'f']
    }


def test_decode_with_transition():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [Reify(color_repeats_factor)])

    solution = [ 1,  -2,  3,  -4,
                 5,  -6, -7,   8,   #  17, -18
                -9,  10,  11, -12,  # -19,  20
                -13, 14, -15,  16,  #  21, -22
                17, -18, -19, 20, 21, -22] # color_repeats_factor
    shuffle(solution)

    decoded = Gen.decode(block, solution)
    assert decoded['color'] ==          ['red', 'red',  'blue', 'blue']
    assert decoded['text']  ==          ['red', 'blue', 'red',  'blue']
    assert decoded['color repeats?'] == ['',    'yes',  'no',   'yes' ]


    solution = [ 1,  -2,  -3,   4,
                -5,   6,   7,  -8,
                -9,   10, -11,  12,
                 13, -14,  15, -16,
                -17,  18,  19, -20, -21, 22]
    shuffle(solution)

    decoded = Gen.decode(block, solution)
    assert decoded['color'] ==          ['red',  'blue', 'blue', 'red']
    assert decoded['text']  ==          ['blue', 'red',  'blue', 'red']
    assert decoded['color repeats?'] == ['',     'no',   'yes',  'no' ]


def test_decode_with_general_window():
    block = CrossBlock([color, text, congruent_bookend],
                       [color, text],
                       [Reify(congruent_bookend)])

    solution = [ 1,  -2,  -3,   4,
                -5,   6,   7,  -8,
                -9,   10, -11,  12,
                 13, -14,  15, -16,
                -17,  18,  19, -20]
    shuffle(solution)

    decoded = Gen.decode(block, solution)
    assert decoded['color'] ==              ['red',  'blue', 'blue', 'red']
    assert decoded['text']  ==              ['blue', 'red',  'blue', 'red']
    assert decoded['congruent bookend?'] == ['no',   '',     '',     'yes']


def test_decode_with_transition_and_only_positive_variables():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [Reify(color_repeats_factor)])

    solution = [2, 3, 5, 8, 18, 9, 11, 19, 14, 16, 22]

    decoded = Gen.decode(block, solution)
    assert decoded['color'] ==          ['blue', 'red',  'red', 'blue']
    assert decoded['text']  ==          ['red',  'blue', 'red', 'blue']
    assert decoded['color repeats?'] == ['',     'no',   'yes', 'no'  ]



"""This module provides constraints for CNF generation."""

import operator as op
from abc import abstractmethod
from copy import deepcopy
from typing import List, Tuple, Any, Union, cast, Dict, Callable
from itertools import chain, product
from math import ceil

from sweetpea._internal.base_constraint import Constraint
from sweetpea._internal.iter import chunk, chunk_list
from sweetpea._internal.block import Block
from sweetpea._internal.cross_block import MultiCrossBlockRepeat
from sweetpea._internal.backend import LowLevelRequest, BackendRequest
from sweetpea._internal.logic import If, Iff, And, Or, Not
from sweetpea._internal.primitive import DerivedFactor, DerivedLevel, Factor, Level, SimpleLevel
from sweetpea._internal.argcheck import argcheck, make_istuple
from sweetpea._internal.weight import combination_weight
from sweetpea._internal.beforestart import BeforeStart

def validate_factor(block: Block, factor: Factor) -> None:
    if not block.has_factor(factor):
        raise ValueError(("A factor with name '{}' wasn't found in the design. "
                          "Are you sure the factor was included, and that the name is spelled "
                          "correctly?").format(factor.name))

def validate_factor_and_level(block: Block, factor: Factor, level: Union[SimpleLevel, DerivedLevel]) -> None:
    validate_factor(block, factor)

    if not level in factor:
        raise ValueError(("A level with name '{}' wasn't found in the '{}' factor").format(
                              level.name,
                              factor.name))


class Consistency(Constraint):
    """This constraint ensures that only one level of each factor is 'on' at a
    time. So for instance in the experiment::

        color = Factor("color", ["red", "blue"])
        text  = Factor("text",  ["red", "blue"])
        design = crossing = [color, text, conFactor]
        experiment   = fully_cross_block(design, crossing, [])

    The first trial is represented by the boolean vars ``[1, 2, 3, 4]``:

    - 1 is true iff the trial is color:red
    - 2 is true iff the trial is color:blue
    - 3 is true iff the trial is text:red
    - 4 is true iff the trial is text:blue

    The second trial is represented by the boolean vars ``[5-8]``, the third by
    ``[9-12]``, the fourth by ``[13-16]``. So this desugaring applies the
    following constraints::

        sum(1, 2) EQ 1
        sum(3, 4) EQ 1
        sum(5, 6) EQ 1
        ...
        sum(15, 16) EQ 1
    """

    def validate(self, block: Block) -> None:
        pass

    @staticmethod
    def apply(block: Block, backend_request: BackendRequest) -> None:
        next_var = 1
        for _ in range(block.trials_per_sample()):
            for f in filter(lambda f: not f.has_complex_window, block.act_design):
                number_of_levels = len(f.levels)
                new_request = LowLevelRequest("EQ", 1, list(range(next_var, next_var + number_of_levels)))
                backend_request.ll_requests.append(new_request)
                next_var += number_of_levels

        for f in filter(lambda f: f.has_complex_window, block.act_design):
            variables_for_factor = block.variables_for_factor(f)
            var_list = list(map(lambda n: n + next_var, range(variables_for_factor)))
            chunks = list(chunk_list(var_list, len(f.levels)))
            backend_request.ll_requests += list(map(lambda v: LowLevelRequest("EQ", 1, v), chunks))
            next_var += variables_for_factor

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        # conformance by construction in combinatoric
        return True

class Cross(Constraint):
    """We represent the fully crossed constraint by allocating additional
    boolean variables to represent each unique state. Only factors in crossing
    will contribute to the number of states (there may be factors in the design
    that aren't in the crossing).

    Continuing with the example from :class:`.Consistency`, we will represent
    the states::

        (color:red, text:red)
        (color:red, text:blue)
        (color:blue, text:red)
        (color:blue, text:blue)

    The steps taken are:

    1. Generate intermediate vars

        Using the fresh var counter, allocate ``numTrials * num_states`` new
        vars

    2. Entangle them with block vars

        Add to the CNF queue: ``toCNF(Iff(newVar, And(levels)))``, e.g., if the
        variable ``1`` indicates ``color:red``, the var ``3`` indicates
        ``text:red``, and the var ``25`` represents ``(color:red, text:red)``,
        do ``toCNF(Iff(25, And([1, 3])))``

    3. 1 hot the *states* e.g., 1 red circle, etc

        Same as :class:`.Consistency` above, collect all the state vars that
        represent each state & enforce that only one of those states is true,
        e.g., ``sum(25, 29, 33, 37) EQ 1`` (and 3 more of these for each of the
        other states).
    """

    def validate(self, block: Block) -> None:
        pass

    @staticmethod
    def apply(block: MultiCrossBlockRepeat, backend_request: BackendRequest) -> None:
        # Treat each crossing seperately, but they're related by shared variables, which
        # are the per-trial, per-level variables of factors used in multiple crossings
        for c in block.crossings:
            fresh = backend_request.fresh

            crossing_size = block.crossing_size(c)
            preamble_size = block.preamble_size(c)
            crossing_weight = block.crossing_weight(c)

            # Step 1a: Get a list of the trials that are involved in the crossing. That list
            # omits leading trials that will be present to initialize transitions, and the
            # number of trials may have been reduced by exclusions.
            crossing_trials = list(filter(lambda t: all(map(lambda f: f.applies_to_trial(t), c)),
                                          range(1, block.trials_per_sample() + 1)))

            # Step 1b: For each trial, cross all levels of all factors in the crossing.
            # We exclude any combination that is dsiallowed by implicit or explicit exlcusions.
            level_lists = [list(f.levels) for f in c]
            crossings = [{level.factor: level for level in levels} for levels in product(*level_lists)]
            trial_combinations = list(filter(lambda c: not block.is_excluded_or_inconsistent_combination(c), crossings))
            crossing_combinations = [[block.encode_combination(c, t) for c in trial_combinations] for t in crossing_trials]
            # Each trial is now represented in `crossing_factors` by a list
            # of potential level combinations, where each level combination is represented
            # as tuple of CNF variables.

            # Step 2a: Allocate additional variables to represent each crossing in each trial.
            num_state_vars = len(crossing_combinations) * len(crossing_combinations[0])
            state_vars = list(range(fresh, fresh + num_state_vars))
            fresh += num_state_vars

            # Step 2b: Associate each state variable with its combination in each trial.
            flattened_combinations = list(chain.from_iterable(crossing_combinations))
            iffs = list(map(lambda n: Iff(state_vars[n], And([*flattened_combinations[n]])), range(len(state_vars))))

            # Step 2c: Get weight associated with each combination.
            combination_weights = [combination_weight(tuple(c.values())) for c in trial_combinations]

            # Step 3: Constrain each crossing to occur exactly according to its weight time the
            # crossing weight in each `crossing_size * crossing_weight` set of trials, or at most
            # that much in a last set of trials that is less than `crossing_size * crossing_weight`
            # in length.
            states = list(chunk(state_vars, len(trial_combinations)))
            transposed = cast(List[List[int]], list(map(list, zip(*states))))
            reqss = map(lambda l, w: Cross.__add_weight_constraint(l, w, crossing_size, crossing_weight),
                        transposed,
                        combination_weights)
            backend_request.ll_requests += list(chain.from_iterable(reqss))

            (cnf, new_fresh) = block.cnf_fn(And(iffs), fresh)

            backend_request.cnfs.append(cnf)
            backend_request.fresh = new_fresh

    @staticmethod
    def __add_weight_constraint(variables: List[int],
                                weight: int,
                                crossing_size: int,
                                crossing_weight: int) -> List[LowLevelRequest]:
        """Constrain to a weight of each `crossing_size` sequence of variables, and at
        at most one for an ending sequence that is less than `crossing_size` in length.
        """
        to_add = len(variables)
        reqs = cast(List[LowLevelRequest], [])
        while to_add > 0:
            if (to_add >= (crossing_size * crossing_weight)):
                reqs.append(LowLevelRequest("EQ", weight*crossing_weight, variables[:(crossing_size*crossing_weight)]))
            else:
                reqs.append(LowLevelRequest("LT", weight*crossing_weight+1, variables))
            variables = variables[(crossing_size*crossing_weight):]
            to_add -= crossing_size * crossing_weight
        return reqs

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        # conformance by construction or direct checking in combinatoric
        return True

class Derivation(Constraint):
    """A derivation such as::

        Derivation(4, [[0, 2], [1, 3]])

    where the index of the derived level is ``4``, and ``[[0, 2], [1, 3]]`` is
    the list of dependent indices, represents the logical formula::

        4 iff (0 and 2) or (1 and 3)

    These indicies are used the get the corresponding trial variables.
    Continuing from the example in of processDerivations, the first trial is
    represented by variables ``[1-6]`` (notice this feels like an off-by-one:
    the indicies start from ``0``, but the boolean variables start from ``1``).
    So we would use the indices to map onto the vars as::

        5 iff (1 and 3) or (2 and 4)

    Then we convert to CNF directly, i.e.::

        toCNF(Iff(5, Or(And(1,3), And(2,4))))

    This is then done for all window-sizes, taking into account strides (which
    are specified only in :class:`DerivedLevels <.DerivedLevel>` specified with
    a general :class:`.Window` rather than :class:`.Transition` or
    :class:`.WithinTrial`). We grab window-sized chunks of the variables that
    represent the trials, map the variables using the indices, and then convert
    to CNF. These chunks look like::

        window1: 1  2  3  4  5  6
        window2: 7  8  9  10 11 12

    So, for the second trial (since the window size in this example is ``1``)
    it would be::

        11 iff (7 and 9) or (8 and 10)

    When a dependent_idx has `BeforeStart`, then it should only apply early
    where the corresponding level is not available.

    90% sure this is the correct way to generalize to derivations involving 2+
    levels & various windowsizes. One test is the experiment::

        color = ["r", "b", "g"];
        text = ["r", "b"];
        conFactor;
        fullycross(color, text) + AtMostKInARow 1 conLevel
    """

    def __init__(self,
                 derived_idx: int,
                 dependent_idxs: List[List[object]],
                 factor: DerivedFactor) -> None:
        self.derived_idx = derived_idx
        self.dependent_idxs = dependent_idxs
        self.factor = factor
        # TODO: validation

    def validate(self, block: Block) -> None:
        pass

    def apply(self, block: Block, backend_request: BackendRequest) -> None:
        if self.is_complex(block):
            self.__apply_derivation(block, backend_request)
        else:
            # If the index is beyond the grid variables, that means it's a derivation from a complex window.
            # (This is brittle, but I haven't come up with a better way yet.)
            self.__apply_derivation_with_complex_window(block, backend_request)

    def is_complex(self, block: Block):
        return self.derived_idx < block.grid_variables()

    def __apply_derivation(self, block: Block, backend_request: BackendRequest) -> None:
        trial_size = block.variables_per_trial()
        cross_size = block.trials_per_sample()

        iffs = []
        for n in range(cross_size):
            or_clause = Or(list(And(list(map(lambda x: x + (n * trial_size) + 1, l))) for l in self.dependent_idxs))
            iffs.append(Iff(self.derived_idx + (n * trial_size) + 1, or_clause))

        (cnf, new_fresh) = block.cnf_fn(And(iffs), backend_request.fresh)

        backend_request.cnfs.append(cnf)
        backend_request.fresh = new_fresh

    def __apply_derivation_with_complex_window(self, block: Block, backend_request: BackendRequest) -> None:
        trial_size = block.variables_per_trial()
        trial_count = block.trials_per_sample()
        iffs = []
        f = self.factor
        window = f.levels[0].window
        t = 0
        delta = window.start_delta
        for n in range(trial_count):
            if not f.applies_to_trial(n + 1):
                continue
            num_levels = len(f.levels)
            get_trial_size = lambda x: trial_size if x < block.grid_variables() else len(block.decode_variable(x+1)[0].levels)

            # Only keep clauses where all `BeforeStarts` apply and all indices are in range:
            ands = []
            for l in self.dependent_idxs:
                vars = cast(List[int], [])
                ok = True
                for x in l:
                    if isinstance(x, BeforeStart):
                        if x.ready_at <= n:
                            ok = False
                            break
                    else:
                        new_x = x + ((t + delta) * window.stride * get_trial_size(x) + 1)
                        if new_x <= 0:
                            ok = False
                            break
                        vars.append(new_x)
                if ok:
                    ands.append(And(vars))

            or_clause = Or(ands)
            iffs.append(Iff(self.derived_idx + (t * num_levels) + 1, or_clause))
            t += 1
        (cnf, new_fresh) = block.cnf_fn(And(iffs), backend_request.fresh)

        backend_request.cnfs.append(cnf)
        backend_request.fresh = new_fresh

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def uses_factor(self, f: Factor) -> bool:
        return any(list(map(lambda l: l.uses_factor(f), self.factor.levels)))

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        return True


class _KInARow(Constraint):
    def __init__(self, k, level):
        self.k = k
        self.level = level
        self.within_block = False
        self.__validate()

    def __validate(self) -> None:
        who = self.__class__.__name__

        if not isinstance(self.k, int):
            raise ValueError(f"{who}: k must be an integer, received {self.k}")

        if self.k <= 0:
            raise ValueError(f"{who}: k must be greater than 0; if you're trying to exclude a particular level, "
                             f"use the 'Exclude' constraint")

        self.level = filter_level(who, self.level, True)

    def validate(self, block: Block) -> None:
        validate_factor_and_level(block, self.level.get_factor(), self.level)

    def set_within_block(self) -> None:
        self.within_block = True

    def uses_factor(self, f: Factor) -> bool:
        if isinstance(self.level, Factor):
            return self.level.uses_factor(f)
        else:
            return self.level.factor.uses_factor(f)

    def desugar(self, replacements: dict) -> List[Constraint]:
        constraints = cast(List[Constraint], [self])

        level = replacements.get(self.level, self.level)

        # Generate the constraint for each level in the factor.
        if isinstance(level, Factor):
            levels = level.levels  # Get the actual levels out of the factor.

            constraints = []
            for l in levels:
                constraint_copy = deepcopy(self)
                constraint_copy.level = l
                constraints.append(constraint_copy)
        elif level != self.level:
            constraint_copy = deepcopy(self)
            constraint_copy.level = level
            constraints = [constraint_copy]

        return constraints

    def apply(self, block: Block, backend_request: BackendRequest) -> None:
        # By this point, level should be a level tht has a factor.
        # Block construction is expected to flatten out constraints applied to whole factors so
        # that the constraint is applied to each level of the factor.
        self.apply_to_backend_request(block, (self.level.factor, self.level), backend_request)

    def _build_variable_sublists(self, block: Block, level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]], sublist_length: int) -> List[List[int]]:
        var_lists = block.build_variable_lists(level, self.within_block)
        sublists = cast(List[List[int]], [])
        for var_list in var_lists:
            raw_sublists = [var_list[i:i+sublist_length] for i in range(0, len(var_list))]
            sublists = sublists + list(filter(lambda l: len(l) == sublist_length, raw_sublists))
        return sublists

    @abstractmethod
    def apply_to_backend_request(self, block: Block, level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]], backend_request: BackendRequest) -> None:
        pass

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        level = self.level
        factor = level.factor
        level_list = sample[factor]

        counts = []
        count = 0
        for l in level_list:
            if count > 0 and l != level:
                counts.append(count)
                count = 0
            elif l == level:
                count += 1
        if count > 0:
            counts.append(count)

        return self._potential_counts_conform(counts)

    @abstractmethod
    def _potential_counts_conform(self, counts: List[int]) -> bool:
        pass

    def _potential_counts_conform_individually(self, counts: List[int], fn: Callable[[int, int], bool]) -> bool:
        return all(map(lambda n: fn(n, self.k), counts))


class AtMostKInARow(_KInARow):
    """This desugars pretty directly into the llrequests. The only thing to do
    here is to collect all the boolean vars that match the same level & pair
    them up according to k.

    Continuing with the example from :class:`.Consistency`, say we want
    ``AtMostKInARow 1 ("color", "red")``, then we need to grab all the vars
    which indicate color-red::

        [1, 7, 13, 19]

    and then wrap them up so that we're making requests like::

        sum(1, 7)  LT 2
        sum(7, 13)  LT 2
        sum(13, 19) LT 2

    If it had been ``AtMostKInARow 2 ("color", "red")``, the reqs would have
    been::

        sum(1, 7, 13)  LT 3
        sum(7, 13, 19) LT 3
    """
    def apply_to_backend_request(self, block: Block, level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]], backend_request: BackendRequest) -> None:
        sublists = self._build_variable_sublists(block, level, self.k + 1)

        # Build the requests
        backend_request.ll_requests += list(map(lambda l: LowLevelRequest("LT", self.k + 1, l), sublists))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def _potential_counts_conform(self, counts: List[int]) -> bool:
        return self._potential_counts_conform_individually(counts, op.le)


class AtLeastKInARow(_KInARow):
    """This is more complicated that AtMostKInARow. We collect all the boolean
    vars that match the same level & pair them up according to k.

    We want ``AtLeastKInARow 2 ("color", "red")``, then we need to grab all the
    vars which indicate color-red::

        [1, 7, 13, 19]

    and then wrap them up in CNF as follows::

        If(1) Then (7)          --------This is a corner case
        If(And(!1, 7)) Then (13)
        If(And(!7, 13)) Then (19)
        If(19) Then (13)   --------This is a corner case

    If it had been ``AtLeastKInARow 3 ("color", "red")``, the CNF would have
    been::

        If(1) Then (7, 13)          --------This is a corner case
        If(And(!1, 7)) Then (13, 19)
        If(19) Then (7, 13)   --------This is a corner case
    """
    def __init__(self, k, levels):
        super().__init__(k, levels)
        self.max_trials_required = cast(int, None)

    def apply_to_backend_request(self, block: Block, level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                                    backend_request: BackendRequest) -> None:

        # Request sublists for k+1 to allow us to determine the transition
        sublists = self._build_variable_sublists(block, level, self.k + 1)
        implications = []
        print(sublists)
        if sublists:
            # Starting corner case
            implications.append(If(sublists[0][0], And(sublists[0][1:-1])))
            for sublist in sublists:
                implications.append(If(And([Not(sublist[0]), sublist[1]]), And(sublist[2:])))
            # Ending corner case
            implications.append(If(Not(sublists[-1][1]), Not(Or(sublists[-1][2:]))))

        (cnf, new_fresh) = block.cnf_fn(And(implications), backend_request.fresh)

        backend_request.cnfs.append(cnf)
        backend_request.fresh = new_fresh

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def _potential_counts_conform(self, counts: List[int]) -> bool:
        return self._potential_counts_conform_individually(counts, op.ge)


class ExactlyK(_KInARow):
    """Requires that if the given level exists at all, it must exist in a trial
    exactly ``k`` times.
    """
    def apply_to_backend_request(self,
                                 block: Block,
                                 level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                                 backend_request: BackendRequest
                                 ) -> None:
        sublistss = block.build_variable_lists(level, self.within_block)
        for sublists in sublistss:
            backend_request.ll_requests.append(LowLevelRequest("EQ", self.k, sublists))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def _potential_counts_conform(self, counts: List[int]) -> bool:
        return sum(counts) == self.k


class ExactlyKInARow(_KInARow):
    """Requires that if the given level exists at all, it must exist in a
    sequence of exactly K.
    """
    def apply_to_backend_request(self,
                                 block: Block,
                                 level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                                 backend_request: BackendRequest
                                 ) -> None:
        sublists = self._build_variable_sublists(block, level, self.k)
        implications = []

        # Handle the regular cases (1 => 2 ^ ... ^ n ^ ~n+1)
        trim = len(sublists) if self.k > 1 else len(sublists) - 1
        for idx, l in enumerate(sublists[:trim]):
            if idx > 0:
                p_list = [Not(sublists[idx-1][0]), l[0]]
                p = And(p_list) if len(p_list) > 1 else p_list[0]
            else:
                p = l[0]

            if idx < len(sublists) - 1:
                q_list = cast(List[Any], l[1:]) + [Not(sublists[idx+1][-1])]
                q = And(q_list) if len(q_list) > 1 else q_list[0]
            else:
                q = And(l[1:]) if len(l[1:]) > 1 else l[self.k - 1]
            implications.append(If(p, q))

        # Handle the tail
        if len(sublists[-1]) > 1:
            tail = sublists[-1]
            tail.reverse()
            for idx in range(len(tail) - 1):
                implications.append(If(l[idx], l[idx + 1]))

        (cnf, new_fresh) = block.cnf_fn(And(implications), backend_request.fresh)
        backend_request.cnfs.append(cnf)
        backend_request.fresh = new_fresh

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def _potential_counts_conform(self, counts: List[int]) -> bool:
        return self._potential_counts_conform_individually(counts, op.eq)

def filter_level(who, level, factor_ok: bool = False):
    if factor_ok and isinstance(level, Factor):
        return level
    elif isinstance(level, Level):
        if hasattr(level, 'factor'):
            return level
        raise ValueError(f"{who}: level does not belong to a factor: {level}")
    elif isinstance(level, tuple) and len(level) == 2 and isinstance(level[0], Factor):
        if isinstance(level[1], SimpleLevel) or isinstance(level[1], DerivedLevel):
            if level[1] not in level[0]:
                raise ValueError(f"{who}: level {level[0]} is not in factor {level[1]}")
            return level[1]
        else:
            l = level[0].get_level(level[1])
            if not l:
                raise ValueError(f"{who}: not a level in factor {level[0]}: {level[1]}")
            return l
    else:
        if factor_ok:
            raise ValueError(f"{who}: expected either a Factor, Level, or a tuple of Factor and Level, given {level}")
        else:
            raise ValueError(f"{who}: expected either a Level or a tuple of Factor and Level, given {level}")

class Exclude(Constraint):
    def __init__(self, level):
        level = filter_level("Exclude", level)
        self.factor = level.factor
        self.level = level

    def validate(self, block: Block) -> None:
        validate_factor_and_level(block, self.factor, self.level)

        block.exclude.append((self.factor, self.level))
        # Store the basic factor-level combnations resulting in the derived excluded factor in the block
        if isinstance(self.level, DerivedLevel) and not self.factor.has_complex_window:
            block.excluded_derived.extend(self.extract_simplelevel(block, self.level))

    def uses_factor(self, f: Factor) -> bool:
        return self.factor.uses_factor(f)

    def desugar(self, replacements: dict) -> List:
        level = replacements.get(self.level, self.level)
        return [Exclude(level)]

    def extract_simplelevel(self, block: Block, level: DerivedLevel) -> List[Dict[Factor, SimpleLevel]]:
        """Recursively deciphers the excluded level to a list of combinations
        basic levels."""
        excluded_levels = []
        excluded: List[Tuple[Level, ...]] = [cross for cross in level.get_dependent_cross_product()
                                             if level.window.predicate(*[level.name for level in cross])]
        for excluded_level_tuple in excluded:
            combos: List[Dict[Factor, SimpleLevel]] = [{}]
            for excluded_level in excluded_level_tuple:
                if isinstance(excluded_level, DerivedLevel):
                    result = self.extract_simplelevel(block, excluded_level)
                    newcombos = []
                    valid = True
                    for r in result:
                        for c in combos:
                            for f in c:
                                if f in r:
                                    if c[f] != r[f]:
                                        valid = False
                        if valid:
                            newcombos.append({**r, **c})
                    combos = newcombos
                else:
                    if not isinstance(excluded_level, SimpleLevel):
                        raise ValueError(f"Unexpected level type in exclusion: level {level.name} of type "
                                         f"{type(level).__name__}.")
                    for c in combos:
                        if block.factor_in_crossing(excluded_level.factor) and block.require_complete_crossing:
                            block.errors.add("WARNING: Some combinations have been excluded, this crossing may not be complete!")
                        c[excluded_level.factor] = excluded_level
            excluded_levels.extend(combos)
        return excluded_levels

    def apply(self, block: Block, backend_request: BackendRequest) -> None:
        var_lists = block.build_variable_lists((self.factor, self.level))
        for var_list in var_lists:
            backend_request.cnfs.append(And(list(map(lambda n: n * -1, var_list))))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def is_complex_for_combinatoric(self) -> bool:
        return False

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        # conformance by construction in combinatoric for simple factors, but
        # we have to check exlcusions based on complex factors
        if self.factor.has_complex_window:
            levels = sample[self.factor]
            level = self.level
            for l in levels:
                if l == level:
                    return False
        return True

class Pin(Constraint):
    def __init__(self, index, level):
        level = filter_level("Pin", level)
        self.index = index
        self.factor = level.factor
        self.level = level
        self.within_block = False

    def set_within_block(self) -> None:
        self.within_block = True

    def validate(self, block: Block) -> None:
        validate_factor_and_level(block, self.factor, self.level)
        if not block.get_trial_numbers(self.index):
            num_trials = block.trials_per_sample()
            block.errors.add("WARNING: Pin constraint unsatisfiable, because "
                             + str(self.index) + " is out of range for " + str(num_trials) + " trials")

    def uses_factor(self, f: Factor) -> bool:
        return self.factor.uses_factor(f)

    def desugar(self, replacements: dict) -> List:
        level = replacements.get(self.level, self.level)
        p = Pin(self.index, level)
        p.within_block = self.within_block
        return [p]

    def apply(self, block: Block, backend_request: BackendRequest) -> None:
        trial_nos = block.get_trial_numbers(self.index, self.within_block)
        if trial_nos:
            for trial_no in trial_nos:
                var = block.get_variable(trial_no+1, (self.factor, self.level))
                backend_request.cnfs.append(And([var]))
        else:
            backend_request.cnfs.append(And([1, -1]))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def is_complex_for_combinatoric(self) -> bool:
        return True

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        levels = sample[self.factor]
        trial_nos = block.get_trial_numbers(self.index, self.within_block)
        if trial_nos:
            for trial_no in trial_nos:
                if levels[trial_no] != self.level:
                    return False
            return True
        else:
            return False

class Reify(Constraint):
    """The only purpose of this constraint is to make a factor
    non-implied, so that it's exposed to a constraint solver."""
    def __init__(self, factor):
        self.factor = factor

    def validate(self, block: Block) -> None:
        validate_factor(block, self.factor)

    def apply(self, block: Block, backend_request: BackendRequest) -> None:
        """Do nothing."""

    def uses_factor(self, f: Factor) -> bool:
        return self.factor.uses_factor(f)

    def is_complex_for_combinatoric(self) -> bool:
        return False

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        return True

    def desugar(self, replacements: dict) -> List:
        factor = replacements.get(self.factor, self.factor)
        return [Reify(factor)]


class MinimumTrials(Constraint):
    def __init__(self, trials):
        self.trials = trials
        who = "MinimumTrials"
        argcheck(who, trials, int, "an integer")
        # TODO: validation

    def is_complex_for_combinatoric(self) -> bool:
        return False

    def validate(self, block: Block) -> None:
        if self.trials <= 0 and not isinstance(self.trials, int):
            raise ValueError("Minimum trials must be a positive integer.")

    def apply(self, block: Block, backend_request: Union[BackendRequest, None]) -> None:
        if block.min_trials:
            block.min_trials = max([block.min_trials, self.trials])
        else:
            block.min_trials = self.trials

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def potential_sample_conforms(self, sample: dict, block: Block) -> bool:
        return True



import pytest
import operator as op

from sweetpea import CrossBlock
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import Reify
from sweetpea._internal.encoding_diagram import __generate_encoding_diagram


color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("color repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

text_repeats_factor = Factor("text repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [text])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [text]))
])

design = [color, text, con_factor]
crossing = [color, text]
blk = CrossBlock(design, crossing, [Reify(con_factor)])


def test_generate_encoding_diagram():
    assert __generate_encoding_diagram(blk) == "\
----------------------------------------------\n\
|   Trial |  color   |   text   | congruent? |\n\
|       # | red blue | red blue |  con  inc  |\n\
----------------------------------------------\n\
|       1 |  1   2   |  3   4   |   5    6   |\n\
|       2 |  7   8   |  9   10  |  11    12  |\n\
|       3 | 13   14  | 15   16  |  17    18  |\n\
|       4 | 19   20  | 21   22  |  23    24  |\n\
----------------------------------------------\n"


def test_generate_encoding_diagram_with_transition():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       [Reify(color_repeats_factor)])

    assert __generate_encoding_diagram(block) == "\
--------------------------------------------------\n\
|   Trial |  color   |   text   | color repeats? |\n\
|       # | red blue | red blue |   yes     no   |\n\
--------------------------------------------------\n\
|       1 |  1   2   |  3   4   |                |\n\
|       2 |  5   6   |  7   8   |   17      18   |\n\
|       3 |  9   10  | 11   12  |   19      20   |\n\
|       4 | 13   14  | 15   16  |   21      22   |\n\
--------------------------------------------------\n"


def test_generate_encoding_diagram_with_constraint_and_multiple_transitions():
    design = [color, text, con_factor, color_repeats_factor, text_repeats_factor]
    block = CrossBlock(design,
                       [color, text],
                       list(map(Reify, design)))

    assert __generate_encoding_diagram(block) == "\
-------------------------------------------------------------------------------\n\
|   Trial |  color   |   text   | congruent? | color repeats? | text repeats? |\n\
|       # | red blue | red blue |  con  inc  |   yes     no   |   yes    no   |\n\
-------------------------------------------------------------------------------\n\
|       1 |  1   2   |  3   4   |   5    6   |                |               |\n\
|       2 |  7   8   |  9   10  |  11    12  |   25      26   |   31     32   |\n\
|       3 | 13   14  | 15   16  |  17    18  |   27      28   |   33     34   |\n\
|       4 | 19   20  | 21   22  |  23    24  |   29      30   |   35     36   |\n\
-------------------------------------------------------------------------------\n"


def test_generate_encoding_diagram_with_constraint_and_multiple_transitions_in_different_order():
    design = [text_repeats_factor, color, color_repeats_factor, text, con_factor]
    block = CrossBlock(design,
                              [color, text],
                              list(map(Reify, design)))

    assert __generate_encoding_diagram(block) == "\
-------------------------------------------------------------------------------\n\
|   Trial | text repeats? |  color   | color repeats? |   text   | congruent? |\n\
|       # |   yes    no   | red blue |   yes     no   | red blue |  con  inc  |\n\
-------------------------------------------------------------------------------\n\
|       1 |               |  1   2   |                |  3   4   |   5    6   |\n\
|       2 |   25     26   |  7   8   |   31      32   |  9   10  |  11    12  |\n\
|       3 |   27     28   | 13   14  |   33      34   | 15   16  |  17    18  |\n\
|       4 |   29     30   | 19   20  |   35      36   | 21   22  |  23    24  |\n\
-------------------------------------------------------------------------------\n"


def test_generate_encoding_diagram_with_windows():
    color3 = Factor("color3", ["red", "blue", "green"])

    yes_fn = lambda colors: colors[0] == colors[-1] == colors[-2]
    no_fn = lambda colors: not yes_fn(colors)
    color3_repeats_factor = Factor("color3 repeats?", [
        DerivedLevel("yes", Window(yes_fn, [color3], 3, 1)),
        DerivedLevel("no",  Window(no_fn, [color3], 3, 1))
    ])

    block = CrossBlock([color3_repeats_factor, color3, text], [color3, text], [Reify(color3_repeats_factor)])

    assert __generate_encoding_diagram(block) == "\
---------------------------------------------------------\n\
|   Trial | color3 repeats? |     color3     |   text   |\n\
|       # |   yes      no   | red blue green | red blue |\n\
---------------------------------------------------------\n\
|       1 |                 |  1   2     3   |  4   5   |\n\
|       2 |                 |  6   7     8   |  9   10  |\n\
|       3 |    31      32   | 11   12   13   | 14   15  |\n\
|       4 |    33      34   | 16   17   18   | 19   20  |\n\
|       5 |    35      36   | 21   22   23   | 24   25  |\n\
|       6 |    37      38   | 26   27   28   | 29   30  |\n\
---------------------------------------------------------\n"


def test_generate_encoding_diagram_with_window_with_stride():
    congruent_bookend = Factor("congruent bookend?", [
        DerivedLevel("yes", Window(lambda colors, texts: colors[0] == texts[0], [color, text], 1, 3)),
        DerivedLevel("no",  Window(lambda colors, texts: colors[0] != texts[0], [color, text], 1, 3))
    ])

    block = CrossBlock([color, text, congruent_bookend], [color, text], [Reify(congruent_bookend)])

    assert __generate_encoding_diagram(block) == "\
------------------------------------------------------\n\
|   Trial |  color   |   text   | congruent bookend? |\n\
|       # | red blue | red blue |    yes       no    |\n\
------------------------------------------------------\n\
|       1 |  1   2   |  3   4   |    17        18    |\n\
|       2 |  5   6   |  7   8   |                    |\n\
|       3 |  9   10  | 11   12  |                    |\n\
|       4 | 13   14  | 15   16  |    19        20    |\n\
------------------------------------------------------\n"

    congruent_bookend = Factor("congruent bookend?", [
        DerivedLevel("yes", Window(lambda colors, texts: colors[0] == texts[0], [color, text], 2, 2)),
        DerivedLevel("no",  Window(lambda colors, texts: colors[0] != texts[0], [color, text], 2, 2))
    ])

    block = CrossBlock([color, text, congruent_bookend], [color, text], [Reify(congruent_bookend)])

    assert __generate_encoding_diagram(block) == "\
------------------------------------------------------\n\
|   Trial |  color   |   text   | congruent bookend? |\n\
|       # | red blue | red blue |    yes       no    |\n\
------------------------------------------------------\n\
|       1 |  1   2   |  3   4   |                    |\n\
|       2 |  5   6   |  7   8   |    17        18    |\n\
|       3 |  9   10  | 11   12  |                    |\n\
|       4 | 13   14  | 15   16  |    19        20    |\n\
------------------------------------------------------\n"




"""Provides a simple custom sequence type for internal use."""

from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from typing import Dict, Iterable, List, MutableSequence, Type, TypeVar, Union, overload


__all__ = ['SimpleSequence']


#: A generic type.
_T = TypeVar('_T')


class SimpleSequence(MutableSequence[_T]):
    """A custom generic sequence. :class:`SimpleSequence` features an advanced
    initialization mechanism that allows for automatically converting given
    arguments into elements of the desired type. It can also smartly handle
    being given lists of elements for initialization.

    The purpose of :class:`SimpleSequence` was to grant classes like
    :class:`Clause` and :class:`CNF` the ability to be initialized with
    literals without needing to duplicate that code. :class:`SimpleSequence` is
    a :class:`typing.MutableSequence` with all the expected functionality, and
    it also provides built-in support for :func:`copy.copy` and
    :func:`copy.deepcopy`.

    :class:`SimpleSequence` was also built to support better type-checking
    throughout SweetPea. To that effect, the implementation was inspired by
    discussion in `mypy issue #4108
    <https://github.com/python/mypy/issues/4108>`_.
    """

    _vals: List[_T]

    @classmethod
    @abstractmethod
    def _get_element_type(cls) -> Type[_T]:
        """Returns the class corresponding to the element type, from which new
        elements can be constructed.
        """
        raise NotImplementedError()

    @classmethod
    def _construct_element(cls, value) -> _T:
        if isinstance(value, cls._get_element_type()):
            return value
        # NOTE: mypy reports "Too many arguments for 'object'" on the following
        #       line. However, the code works as expected, so typechecking is
        #       disabled for this line.
        return cls._get_element_type()(value)  # type: ignore

    def __init__(self, first_value: Union[None, List[Union[_T, int]], _T, int] = None, *rest_values: Union[_T, int]):
        values: Iterable[Union[_T, int]]
        if first_value is None:
            if rest_values:
                raise ValueError(f"cannot instantiate {type(self).__name__} with both None and variadic arguments")
            values = []
        elif isinstance(first_value, (list, tuple)):
            if rest_values:
                raise ValueError(f"cannot instantiate {type(self).__name__} with both list and variadic arguments")
            values = first_value
        else:
            values = first_value, *rest_values
        self._vals = [self._construct_element(value) for value in values]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join(map(repr, self._vals))})"

    def __copy__(self) -> SimpleSequence[_T]:
        new_sequence = self.__class__()
        new_sequence._vals = self._vals
        return new_sequence

    def __deepcopy__(self, memo: Dict) -> SimpleSequence[_T]:
        new_base_vals = deepcopy(self._vals, memo)
        new_sequence = self.__class__()
        new_sequence._vals = new_base_vals
        return new_sequence

    @overload
    def __getitem__(self, index: int) -> _T:
        pass

    @overload
    def __getitem__(self, index: slice) -> SimpleSequence[_T]:
        pass

    def __getitem__(self, index: Union[int, slice]) -> Union[_T, SimpleSequence[_T]]:
        if isinstance(index, slice):
            return self.__class__(*self._vals[index])
        return self._vals[index]

    @overload
    def __setitem__(self, index: int, item: _T) -> None:
        pass

    @overload
    def __setitem__(self, index: slice, item: Iterable[_T]) -> None:
        pass

    def __setitem__(self, index: Union[int, slice], item: Union[_T, Iterable[_T]]) -> None:
        # FIXME: As of mypy 0.790 with Python 3.9.1, mypy is unable to handle
        #        the separate overloads for this function as documented in:
        #            https://github.com/python/mypy/issues/7858
        #        If this is fixed, remove the following pragma.
        self._vals[index] = item  # type: ignore

    @overload
    def __delitem__(self, index: int) -> None:
        pass

    @overload
    def __delitem__(self, index: slice) -> None:
        pass

    def __delitem__(self, index: Union[int, slice]) -> None:
        self._vals.__delitem__(index)

    def insert(self, index: int, item: _T) -> None:
        """Inserts the ``item`` before the given ``index`` in the sequence."""
        self._vals.insert(index, item)

    def __len__(self) -> int:
        return len(self._vals)



"""This module provides functionality for calling the third-party CryptoMiniSAT
tool.

`CryptoMiniSAT <https://github.com/msoos/cryptominisat>`_ is a an advanced
incremental SAT solver. SweetPea uses CryptoMiniSAT for a few processes,
including solving some CNF formulas or checking whether a CNF formula is
satisfiable to begin with.
"""


from pathlib import Path
from shlex import split as shell_split
from subprocess import CompletedProcess, run
from typing import List, Optional, Tuple

from .docker_utility import DEFAULT_DOCKER_MODE_ON, docker_run
from .executables import CRYPTOMINISAT_EXE, DEFAULT_DOWNLOAD_IF_MISSING, ensure_executable_available
from .return_code import ReturnCodeEnum
from .tool_error import ToolError


__all__ = ['DEFAULT_DOCKER_MODE_ON', 'cryptominisat_solve', 'cryptominisat_is_satisfiable']


class CryptoMiniSATReturnCode(ReturnCodeEnum):
    """CryptoMiniSAT uses some unconventional return codes to indicate the
    result of various computations.
    """

    #: CryptoMiniSAT judged the CNF formula to be satisfiable.
    Satisfiable   = 10
    #: CryptoMiniSAT can make no judgment about the CNF formula.
    Unknown       = 15
    #: CryptoMiniSAT judged the CNF formula to be unsatisfiable.
    Unsatisfiable = 20


class CryptoMiniSATError(ToolError):
    """An error raised when CryptoMiniSAT fails."""
    pass


def call_cryptominisat_docker(input_file: Path) -> CompletedProcess:
    """Calls CryptoMiniSAT in a Docker container, reading a given file as the
    input problem.
    """
    cms_container = 'msoos/cryptominisat'
    input_bytes = input_file.read_bytes()
    args = shell_split("--rm -i -a stdin -a stdout")
    result = docker_run(cms_container, args, input_bytes)
    return result


def call_cryptominisat_cli(input_file: Path, download_if_missing: bool) -> CompletedProcess:
    """Calls CryptoMiniSAT from the command line, reading a given file as the
    input problem.

    If ``download_if_missing`` is ``True``, SweetPea will automatically
    download the CryptoMiniSAT executable (and other executables SweetPea
    depends on) to a local directory from the `sweetpea-org/unigen-exe
    repository <https://github.com/sweetpea-org/unigen-exe>`_.
    """
    ensure_executable_available(CRYPTOMINISAT_EXE, download_if_missing)
    command = [str(CRYPTOMINISAT_EXE), "--verb=0", str(input_file)]
    result = run(command, capture_output=True)
    return result


def call_cryptominisat(input_file: Path,
                       docker_mode: bool = DEFAULT_DOCKER_MODE_ON,
                       download_if_missing: bool = DEFAULT_DOWNLOAD_IF_MISSING
                       ) -> Tuple[str, CryptoMiniSATReturnCode]:
    """Calls CryptoMiniSAT with the given file as input.

    If ``docker_mode`` is ``True``, this will use a Docker container to run
    CryptoMiniSAT. If it's ``False``, a command-line executable will be used.

    If ``docker_mode`` is ``False`` *and* no local CryptoMiniSAT executable can
    be found, and if ``download_if_missing`` is ``True``, the needed executable
    will be automatically downloaded if it's missing.
    """
    if docker_mode:
        result = call_cryptominisat_docker(input_file)
    else:
        result = call_cryptominisat_cli(input_file, download_if_missing)
    if CryptoMiniSATReturnCode.has_value(result.returncode):
        return (result.stdout.decode(), CryptoMiniSATReturnCode(result.returncode))
    else:
        stdout = result.stdout.decode()
        stderr = result.stderr.decode()
        raise CryptoMiniSATError(result.returncode, stdout, stderr)


def cryptominisat_solve(input_file: Path, docker_mode: bool = DEFAULT_DOCKER_MODE_ON) -> Optional[List[int]]:
    """Attempts to solve a CNF formula with CryptoMiniSAT and returns the
    result as a list of integers.

    Returns an empty list if the result was unsatisfiable, and returns ``None``
    if CryptoMiniSAT encounters some unknown issue.
    """
    (result, code) = call_cryptominisat(input_file, docker_mode)
    if code is CryptoMiniSATReturnCode.Unsatisfiable:
        return []
    elif code is CryptoMiniSATReturnCode.Satisfiable:
        parts = ''.join(line for line in map(str.strip, result.strip().splitlines())
                        if line.startswith('v')).replace('v', '').split()
        return [int(p) for p in parts]
    else:
        return None


def cryptominisat_is_satisfiable(input_file: Path, docker_mode: bool = DEFAULT_DOCKER_MODE_ON) -> Optional[bool]:
    """Determines whether the CNF formula encoded in the input file is
    satisfiable, according to CryptoMiniSAT.

    Returns ``None`` if CryptoMiniSAT encounters an unknown issue.
    """
    (_, code) = call_cryptominisat(input_file, docker_mode)
    if code is CryptoMiniSATReturnCode.Satisfiable:
        return True
    elif code is CryptoMiniSATReturnCode.Unsatisfiable:
        return False
    else:
        return None



"""The :mod:`sweetpea` module provides the SweetPea domain-specific programming
language. :ref:`The SweetPea language <home>` allows for the specification of
randomized experimental designs and the synthesis of trial sequences from those
designs.
"""

from sweetpea._internal.main import *



"""This module provides the various kinds of blocks that can be used to create
a factorial experimental design.
"""

from abc import abstractmethod
from functools import reduce
from itertools import accumulate, combinations, product, repeat
from typing import List, Union, Tuple, Optional, cast, Any, Dict, Set, Callable, TypeVar
from math import ceil
from networkx import has_path

from sweetpea._internal.backend import BackendRequest
from sweetpea._internal.level import get_all_levels
from sweetpea._internal.primitive import (
    DerivedFactor, DerivedLevel, ElseLevel, Factor, SimpleLevel, Level
)
from sweetpea._internal.logic import to_cnf_tseitin
from sweetpea._internal.base_constraint import Constraint
from sweetpea._internal.design_graph import DesignGraph
from sweetpea._internal.iter import chunk_dict
from sweetpea._internal.weight import combination_weight
from sweetpea._internal.argcheck import argcheck, make_islistof

T = TypeVar('T')
            
class Block:
    """Abstract class for Blocks. Contains the required data, and defines
    abstract methods that other blocks _must_ implement in order to work
    correctly.
    """

    def __init__(self,
                 design: List[Factor],
                 crossings: List[List[Factor]],
                 constraints: List[Constraint],
                 require_complete_crossing,
                 who: str) -> None:
        self.design = list(design).copy()
        self.crossings = list(map(lambda c: list(c).copy(), crossings))
        self.constraints = list(constraints).copy()
        self.cnf_fn = to_cnf_tseitin
        self.complex_factors_or_constraints = True
        self.min_trials = cast(int, 0)
        self.exclude = cast(List[Tuple[Factor, Union[SimpleLevel, DerivedLevel]]], [])
        self.excluded_derived = cast(List[Dict[Factor, SimpleLevel]], [])
        self.require_complete_crossing = require_complete_crossing
        self.errors = cast(Set[str], set())
        self.act_design = list(filter(lambda f: not self.factor_is_implied(f), self.design))
        self._trials_per_sample = None
        self.within_block_count = cast(Optional[int], None)
        self.within_block_preamble = cast(Optional[int], None)
        self._simple_tuples = cast(Optional[List[Tuple[Factor, Union[SimpleLevel, DerivedLevel]]]], None)
        self._variables_per_trial = None
        self.__validate(who)
        self._cached_previous_count = cast(Dict[Tuple[Factor, int], int], {})

    def show_errors(self) -> bool:
        failed = False
        if self.errors:
            for e in self.errors:
                if "WARNING" not in e:
                    print(e)
                    failed = True
            if not failed:
                for e in self.errors:
                    if "WARNING" in e:
                        print(e)
        return failed

    def extract_basic_factor_names(self, level: DerivedLevel) -> set:
        included_factor_names = set()
        for factor in level.window.factors:
            if isinstance(factor, DerivedFactor):
                for l in factor.levels:
                    included_factor_names.update(self.extract_basic_factor_names(l))
            else:
                included_factor_names.add(factor.name)
        return included_factor_names

    def __validate(self, who: str):
        for cr in self.crossings:
            names = set()
            for f in cr:
                if not f in self.design:
                    raise RuntimeError(f"{who}: factor in crossing is not in design: {f}")
                if f.name in names:
                    raise RuntimeError(f"{who}: multiple factors have the same name: {f.name}")
                if f.has_complex_window and cast(DerivedLevel, f.first_level).window.stride > 1:
                    raise RuntimeError(f"{who}: factor with stride > 1 not allowed in crossing: {f.name}")
                names.add(f.name)
        # Ensure basic factors of derived levels in design form complete subset
        # of all declared basic factors.
        basic_factor_names = set()
        included_factor_names = set()
        for design_factor in self.design:
            for level in design_factor.levels:
                if isinstance(level, SimpleLevel):
                    basic_factor_names.add(level.factor.name)
                elif isinstance(level, DerivedLevel):
                    included_factor_names.update(self.extract_basic_factor_names(level))
                else:
                    raise RuntimeError('Expected SimpleLevel or DerivedLevel but found ' + str(type(level)))
        undefined_factor_names = included_factor_names - basic_factor_names
        if undefined_factor_names:
            # The derived levels include factors that are not basic factors.
            raise RuntimeError(f"Derived levels in experiment design include factors that are not listed as basic "
                               f"factors: {', '.join(str(name) for name in undefined_factor_names)}.")
        from sweetpea._internal.constraint import AtLeastKInARow, MinimumTrials
        for c in self.constraints:
            if isinstance(c, MinimumTrials):
                c.apply(self, None)
        for c in self.constraints:
            c.validate(self)
        for c in self.constraints:
            if isinstance(c, AtLeastKInARow):
                c.max_trials_required = self.trials_per_sample() * c.k

    @abstractmethod
    def trials_per_sample(self):
        """Indicates the number of trials that are generated per sample for
        this block configuration.

        Analogous to the old ``__fully_cross_size`` function.
        """
        pass

    @abstractmethod
    def variables_per_trial(self):
        """Indicates the number of variables that are present in each trial.

        Analogous to the old ``__design_size`` function.
        """
        pass

    @abstractmethod
    def grid_variables(self):
        """Indicates the number of variables that are present in the core
        variable grid. this does not include variables used to encode complex
        windows.

        In a design *without* complex windows, this is equivalent to
        :func:`.Block.variables_per_sample`.
        """
        pass

    def variables_per_sample(self):
        """Indicates the total number of variables needed to encode the core
        experiment description.

        Alternatively stated, this returns the number of variables in the
        formula that constitute the independent support.
        """
        return reduce(lambda sum, f: sum + self.variables_for_factor(f), self.act_design, 0)

    def support_variables(self) -> List[int]:
        """Returns the variables for all non-derived factors for all trials.
        The values of these variables determine the values of all others."""
        vars = []
        for t in range(self.trials_per_sample()):
            for f in self.act_design:
                if not isinstance(f, DerivedFactor):
                    vars += self.factor_variables_for_trial(f, t + 1)
        return vars

    def variables_for_factor(self, f: Factor,
                             start: int = 0,
                             end: Optional[int] = None) -> int:
        """Indicates the number of variables needed to encode this factor."""
        trial_list = range(1 + start, (end if end else self.trials_per_sample()) + 1)
        return reduce(lambda sum, t: sum + len(f.levels) if f.applies_to_trial(t) else sum, trial_list, 0)

    def has_factor(self, factor: Factor) -> Factor:
        """Determines whether a given factor is in this block."""
        if not isinstance(factor, Factor):
            raise ValueError('Non-factor argument to has_factor.')
        if factor in self.design:
            return factor
        return cast(Factor, None)

    def first_variable_for_level(self, factor: Factor, level: Any) -> int:
        """Returns the first index for this variable in a trial sequence
        representing the given factor and level. (0-based.)
        """
        if not isinstance(level, (SimpleLevel, DerivedLevel)):
            raise ValueError(f"Attempted to find first variable of non-Level object: {level}.")
        if factor.has_complex_window:
            offset = 0
            complex_factors = filter(lambda f: f.has_complex_window, self.act_design)
            for f in complex_factors:
                if f == factor:
                    offset += f.levels.index(level)
                    break
                else:
                    offset += self.variables_for_factor(f)

            return self.grid_variables() + offset

        else:
            simple_factors = list(filter(lambda f: not f.has_complex_window, self.act_design))
            simple_levels = get_all_levels(simple_factors)
            return simple_levels.index((factor, level))

    def _get_previous_trials_variable_count(self, f: Factor, trial: int):
        """The `trial` argument is 1-based."""
        t = trial
        while True:
            maybe_count = cast(Optional[int], 0)
            if t == 1:
                maybe_count = 0
            else:
                key = (f, t)
                maybe_count = self._cached_previous_count.get(key, None)
            if maybe_count is not None:
                count = maybe_count
                while t < trial:
                    if (f.applies_to_trial(t)):
                        count += 1
                    t += 1
                    self._cached_previous_count[(f, t)] = count
                return count
            t -= 1

    def factor_variables_for_trial(self, f: Factor, t: int) -> List[int]:
        """Given a factor and a trial number (1-based) this function will
        return a list of the variables representing the levels of the given
        factor for that trial. The variable list is also 1-based.
        """
        if not f.applies_to_trial(t):
            raise ValueError('Factor does not apply to trial #' + str(t) + ' f=' + str(f))

        previous_trials = self._get_previous_trials_variable_count(f, t)
        # this could be computed once per factor after self.exclude is in place
        initial_sequence = list(map(lambda l: self.first_variable_for_level(f, l),
                                    list(filter(lambda l: (f, l) not in self.exclude,
                                                f.levels))))
        offset = 0
        if f.has_complex_window:
            offset = len(f.levels) * previous_trials
        else:
            offset = self.variables_per_trial() * previous_trials
        return list(map(lambda n: n + offset + 1, initial_sequence))

    def variable_list_for_trial(self, t: int) -> List[List[int]]:
        """Given a trial number (1-based) this function will return a list of
        lists of the variables that pertain to that trial.

        For example, for stroop-2 with a congruency level, this method would
        return the following for trial ``1``::

          [[1, 2], [3, 4], [5, 6]]

        If a transition were involved, and it didn't apply to level one, then
        the factor would have an empty list::

          [[1, 2], [3, 4], []]
        """
        variables = cast(List[List[int]], [])
        for f in self.act_design:
            # Skip factors that don't apply.
            if not f.applies_to_trial(t):
                variables.append([])
                continue

            variables.append(self.factor_variables_for_trial(f, t))

        return variables

    def _encode_variable(self, f: Factor, l: Level, trial: int):
        offset = self.first_variable_for_level(f, l)
        previous_trials = self._get_previous_trials_variable_count(f, trial)
        if f.has_complex_window:
            offset += len(f.levels) * previous_trials
        else:
            offset += self.variables_per_trial() * previous_trials
        return offset + 1

    def encode_combination(self, combination: Dict[Factor, Level], trial: int):
        return tuple([self._encode_variable(f, l, trial) for f, l in combination.items()])

    def decode_variable(self, variable: int) -> Tuple[Factor, Union[SimpleLevel, DerivedLevel]]:
        """Given a variable number from the SAT formula, this method will
        return the associated factor and level name.
        """
        # Shift to zero-based index
        variable -= 1

        if variable < self.grid_variables():
            variable = variable % self.variables_per_trial()
            if not self._simple_tuples:
                simple_factors = list(filter(lambda f: not f.has_complex_window, self.act_design))
                self._simple_tuples = get_all_levels(simple_factors)
            assert self._simple_tuples  # for the type checker
            return self._simple_tuples[variable]
        else:
            complex_factors = list(filter(lambda f: f.has_complex_window, self.act_design))
            for f in complex_factors:
                start = self.first_variable_for_level(f, f.levels[0])
                end = start + self.variables_for_factor(f)
                if variable in range(start, end):
                    tuples = get_all_levels([f])
                    return tuples[(variable - start) % len(f.levels)]

        raise RuntimeError('Unable to find factor/level for variable!')

    def is_excluded_combination(self, di: Dict[Factor, SimpleLevel]) -> bool:
        """Given a combination of levels, reports whether this combination has been excluded,
        either explicitly or implicitly by the definition of a derived level."""
        # Check based on excluded simple levels:
        if any([t[0] in di and di[t[0]] == t[1] for t in self.exclude]):
            return True
        # Check based on excluded derived levels and derived-level definitions:
        if any([all([e[ex_level] == di.get(ex_level, None) for ex_level in e]) for e in self.excluded_derived]):
            return True
        return False

    def is_excluded_or_inconsistent_combination(self, di: Dict[Factor, SimpleLevel]) -> bool:
        """Like extends is_excluded_combination to also check for combinations that are
        inconsistent with derived-factor definitions. Assumes that `di` is based on the
        block's first crossing.
        """
        if self.is_excluded_combination(di):
            return True
        for f in self.crossings[0]:
            if isinstance(f, DerivedFactor) and not f.has_complex_window and f in di:
                l = cast(DerivedLevel, di[f])
                if all([df in di for df in l.window.factors]):
                    args = [di[df].name for df in l.window.factors]
                    if not l.window.predicate(*args):
                        return True
        return False

    def build_backend_request(self) -> BackendRequest:
        """Apply all constraints to build a :class:`.BackendRequest`. Formerly
        known as ``__desugar``.
        """
        fresh = 1 + self.variables_per_sample()
        backend_request = BackendRequest(fresh)

        from sweetpea._internal.constraint import MinimumTrials
        for c in self.constraints:
            if isinstance(c, MinimumTrials):
                continue
            c.apply(self, backend_request)

        return backend_request

    def get_variable(self, trial_number: int, level: Tuple[Factor, Any]) -> int:
        """Given a trial number (1-based), factor, and level, this method will
        return the SAT variable that represents that selection. Only works for
        factors without complex windows at the moment.
        """
        return self._encode_variable(level[0], level[1], trial_number)

    def build_variable_lists(self,
                             level_pair: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                             within_block: bool = False) -> List[List[int]]:
        """Given a specific level (factor + level pair), this method will
        return the list of variables that correspond to that level in each
        trial in the encoding.
        """
        factor = level_pair[0]
        level = level_pair[1]
        if not isinstance(factor, Factor):
            raise ValueError("First element in level argument to variable list builder must be a Factor.")
        if not isinstance(level, (SimpleLevel, DerivedLevel)):
            raise ValueError("Second element in level argument to variable list builder must be a SimpleLevel "
                             "or a DERIVED LEVEL.")
        if factor.has_complex_window:
            return self.__build_complex_variable_lists(level_pair, within_block)
        else:
            return self.__build_simple_variable_lists(level_pair, within_block)

    def factor_in_crossing(self, factor):
        pass

    def factor_used_in_crossing(self, factor):
        pass

    def factor_is_implied(self, f: Factor) -> bool:
        """Determines whether a factor's level selection is completely implied by level selections
        other factors and is not involved directly in any constraints, so it doesn't have to be
        part of the problem that is passed along to a solver.
        """
        if not isinstance(f, DerivedFactor):
            return False
        if self.factor_used_in_crossing(f):
            return False
        return not any(list(map(lambda c: c.uses_factor(f), self.constraints)))

    def add_implied_levels(self, results: dict) -> dict:
        """Given a dictionary for an experiment that maps all non-implied factors to their levels,
        adds level values for implied factors"""
        n = len(list(results.values())[0])
        for f in self.design:
            if f not in self.act_design:
                vals = []
                for i in range(0, n):
                    if f.applies_to_trial(i + 1):
                        for l in f.levels:
                            if isinstance(l, ElseLevel):
                                vals.append(l.name)
                            elif isinstance(l, DerivedLevel):
                                w = l.window
                                args = []
                                for idx, df in enumerate(w.factors):
                                    for j in range(w.width):
                                        shift = w.width - j - 1
                                        if i - shift >= 0:
                                            args.append(results[df.name][i - shift])
                                        else:
                                            args.append(None)
                                if w.width > 1:
                                    args = list(chunk_dict(args, w.width))
                                if w.predicate(*args):
                                    vals.append(l.name)
                            else:
                                raise RuntimeError("unexpected level in implied factor")
                    else:
                        vals.append("")
                results[f.name] = vals
        return results

    def __map_block_trial_ranges(self, within_block: bool, proc: Callable[[int, int], T]) -> List[T]:
        num_trials = self.trials_per_sample()
        if within_block:
            start = 0
            if self.within_block_count and (self.within_block_preamble != None):
                end = self.within_block_count
                step = self.within_block_count - cast(int, self.within_block_preamble)
            else:
                raise RuntimeError("within-block but not in a repeat?!")
        else:
            start = 0
            end = num_trials
            step = num_trials
        lists = cast(List[T], [])
        while start < num_trials:
            lists.append(proc(start, end))
            start += step
            end += step
        return lists

    def __build_simple_variable_lists(self,
                                      level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                                      within_block: bool = False) -> List[List[int]]:
        def get_variables(start: int, end: int) -> List[int]:
            nonlocal level
            first_variable = self.first_variable_for_level(level[0], level[1]) + 1 + start
            design_var_count = self.variables_per_trial()
            # TODO: This should be reworked. It's an accumulating fold where the
            #       folding function's second argument is just thrown away.
            return list(accumulate(repeat(first_variable, end - start),
                                   lambda acc, _: acc + design_var_count))
        return self.__map_block_trial_ranges(within_block, get_variables)

    def __build_complex_variable_lists(self,
                                       level: Tuple[Factor, Union[SimpleLevel, DerivedLevel]],
                                       within_block: bool = False) -> List[List[int]]:
        factor = level[0]
        level_count = len(factor.levels)
        start_idx = self.first_variable_for_level(level[0], level[1]) + 1
        def get_variables(start: int, end: int) -> List[int]:
            nonlocal factor, level_count, start_idx
            n = self.variables_for_factor(factor, start, end) // level_count
            return reduce(lambda l, v: l + [start_idx + ((v + start) * level_count)], range(n), [])
        return self.__map_block_trial_ranges(within_block, get_variables)

    def get_trial_numbers(self, b_trial_no: int, within_block: bool = False) -> List[int]:
        def get_variables(start: int, end: int) -> int:
            nonlocal b_trial_no
            if b_trial_no < 0:
                trial_no = end + b_trial_no
            else:
                trial_no = start + b_trial_no
            if (trial_no >= start) and (trial_no < end):
                return trial_no
            else:
                return -1
        nums = self.__map_block_trial_ranges(within_block, get_variables)
        if all([n >= 0 for n in nums]):
            return nums
        else:
            return []

    def rearrage_samples(self, samples, results):
        pass

    def calculate_samples_required(self, samples):
        pass

    @abstractmethod
    def sample_mismatch_factors(self, sample: dict) -> list:
        pass

    @abstractmethod
    def sample_mismatch_constraints(self, sample: dict) -> list:
        pass

    @abstractmethod
    def sample_mismatch_crossing(self, sample: dict, acceptable_error_per_crossing: int = 0) -> list:
        pass



import operator as op
import pytest

from sweetpea import *

color = Factor("color",  ["red", "blue"])
size = Factor("size",  ["big", "small"])
direction = Factor("direction",  ["up", "down", "left", "right"])

red = color.get_level('red')
blue = color.get_level('blue')
up = direction.get_level('up')
down = direction.get_level('down')
right = direction.get_level('right')
left = direction.get_level('left')

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('constraints_and_solutions',
                         # only way this works is red, blue, blue, red; solutions: 4
                         [[[AtMostKInARow(1, (color, red)), AtLeastKInARow(2, (color, blue))], 4],
                          [[AtMostKInARow(1, (color, red))], 12],
                          [[AtLeastKInARow(2, (color, red))], 12],
                          [[ExactlyKInARow(2, (color, red))], 12],
                          [[ExactlyK(2, (color, red))], 24]])
def test_check_constraints_on_crossing_factor(strategy, constraints_and_solutions):
    constraints = constraints_and_solutions[0]
    solutions = constraints_and_solutions[1]

    design       = [color, size]
    crossing     = [color, size]
    block        = CrossBlock(design, crossing, constraints)

    experiments  = synthesize_trials(block, 100, sampling_strategy=strategy)

    assert len(experiments) == solutions

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
@pytest.mark.parametrize('constraints_and_solutions',
                         [[[ExactlyK(3, up), ExactlyK(1, right)], 96],
                          [[ExactlyK(4, up)], 24],
                          [[ExactlyKInARow(4, up), ExactlyK(4, up)], 24],
                          [[ExactlyKInARow(3, up), ExactlyK(3, up)], 144],
                          [[ExactlyKInARow(1, up),
                            ExactlyK(2,(direction, up)),
                            Exclude(left),
                            ExactlyKInARow(1, down),
                            ExactlyKInARow(1, right)],
                           240]])
def test_check_constraints_on_design_factor(strategy, constraints_and_solutions):
    constraints = constraints_and_solutions[0]
    solutions = constraints_and_solutions[1]

    design       = [color, size, direction]
    crossing     = [color, size]
    block        = CrossBlock(design, crossing, constraints)

    experiments  = synthesize_trials(block, 500, sampling_strategy=strategy)

    assert len(experiments) == solutions



"""This module provides some convenience functions to be used internally."""


from itertools import islice, tee, chain, repeat
from typing import Any, Tuple, List, Dict, Iterator, Iterable


def chunk(it: Iterable[Any], size: int) -> Iterator[Tuple[Any, ...]]:
    """Handy-dandy chunker from SO:
    https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
    """
    # TODO: Add a canary print statement in case the resulting lists are not
    #       all the same length --- that is not the expected behavior (at least
    #       how it's used in desugar_fullycrossed)
    it = iter(it)
    return iter(lambda: tuple(islice(it, size)), ())


def chunk_list(it: Iterable[Any], size: int) -> Iterator[List[Any]]:
    it = iter(it)
    return iter(lambda: list(islice(it, size)), [])

def chunk_dict(it: Iterable[Any], size: int) -> List[Dict[int, Any]]:
    it = iter(it)
    r = []
    while True:
        try:
            r.append({i-size+1: next(it) for i in range(size)})
        except StopIteration:
            break
    return r

def pairwise(iterable):
    """Helper recipe from:
    https://docs.python.org/3/library/itertools.html#itertools-recipes

    ``s -> (s0,s1), (s1,s2), (s2, s3), ...``
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def intersperse(delimiter, seq, repeat_delimiter = 1):
    """Another helper from SO, with modification for repeating the delimiter.
    https://stackoverflow.com/questions/5655708/python-most-elegant-way-to-intersperse-a-list-with-an-element
    """
    nested_list_seq = map(lambda elem: [elem], seq)
    new_seq = zip(repeat([delimiter] * repeat_delimiter), nested_list_seq)
    return chain(*islice(chain.from_iterable(new_seq), 1, None))



import operator as op
import pytest

from itertools import permutations

from sweetpea import Factor, DerivedLevel, WithinTrial, Transition
from sweetpea import AtMostKInARow, ExactlyKInARow, Exclude
from sweetpea import IterateGen
from sweetpea import CrossBlock, synthesize_trials, synthesize_trials

# Basic setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

# Congruent Factor
con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

# Repeated color Factor
repeated_color_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

# Repeated text Factor
repeated_text_factor = Factor("repeated text?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])


@pytest.mark.parametrize('design', permutations([color, text]))
def test_correct_solution_count(design):
    crossing = [color, text]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, con_factor]))
def test_correct_solution_count_with_congruence_factor_but_unconstrained(design):
    crossing = [color, text]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, con_factor]))
def test_correct_solution_count_with_congruence_factor_and_constrained(design):
    crossing = [color, text]
    constraints = [AtMostKInARow(1, (con_factor, "con"))]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 12


@pytest.mark.parametrize('design', permutations([color, text, con_factor]))
def test_correct_solution_count_with_congruence_factor_and_constrained_exactly(design):
    crossing = [color, text]
    constraints = [ExactlyKInARow(2, con_factor)]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 8


@pytest.mark.parametrize('design', permutations([color, text, repeated_color_factor]))
def test_correct_solution_count_with_repeated_color_factor_but_unconstrained(design):
    crossing = [color, text]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, repeated_color_factor]))
def test_correct_solution_count_with_repeated_color_factor_and_constrained(design):
    crossing = [color, text]
    constraints = [AtMostKInARow(1, (repeated_color_factor, "yes"))]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    # With only two colors, there can never be two color repetitions anyways,
    # so the total should still be the same.
    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, repeated_color_factor, repeated_text_factor]))
def test_correct_solution_count_with_repeated_color_and_text_factors_but_unconstrained(design):
    crossing = [color, text]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, repeated_color_factor, repeated_text_factor]))
def test_correct_solution_count_with_repeated_color_and_text_factors_and_constrained(design):
    crossing = [color, text]
    constraints = [
        AtMostKInARow(1, (repeated_color_factor, "yes")),
        AtMostKInARow(1, (repeated_text_factor, "yes"))
    ]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, repeated_color_factor]))
def test_correct_solution_count_with_repeated_color_factor_and_no_repetition_allowed(design):
    crossing = [color, text]
    constraints = [Exclude(repeated_color_factor["yes"])]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 8



# Everything in `__all_` is exported from the `sweetpea` module.

__all__ = [
    'synthesize_trials', 'sample_mismatch_experiment',

    'print_experiments', 'tabulate_experiments',
    'save_experiments_csv', 'experiments_to_tuples',

    'Block', 'CrossBlock', 'MultiCrossBlock', 'Repeat',

    'Factor', 'Level', 'DerivedLevel', 'ElseLevel',

    'Derivation', 'WithinTrial', 'Transition', 'Window',

    'Constraint',
    'Exclude', 'Pin', 'MinimumTrials', 'ExactlyK',
    'AtMostKInARow', 'AtLeastKInARow',
    'ExactlyKInARow',

    'Gen', 'RandomGen', 'IterateSATGen',
    'CMSGen', 'UniGen', 'IterateILPGen',
    'UniformGen', 'IterateGen'
]

from functools import reduce
from typing import Dict, List, Optional, Tuple, Any, Union, cast
from itertools import product
import csv, os

from sweetpea._internal.block import Block
from sweetpea._internal.cross_block import MultiCrossBlockRepeat, MultiCrossBlock, CrossBlock, Repeat
from sweetpea._internal.primitive import (
    Factor, SimpleFactor, DerivedFactor, Level, SimpleLevel, DerivedLevel, ElseLevel,
    Window, WithinTrial, Transition,
    HiddenName
)
from sweetpea._internal.constraint import (
    Consistency, Constraint, Derivation,
    Exclude, Pin, MinimumTrials,
    ExactlyK, AtMostKInARow, AtLeastKInARow, ExactlyKInARow
)
from sweetpea._internal.sampling_strategy.base import Gen
from sweetpea._internal.sampling_strategy.uniform import UniformGen
from sweetpea._internal.sampling_strategy.iterate import IterateGen
from sweetpea._internal.sampling_strategy.iterate_sat import IterateSATGen
from sweetpea._internal.sampling_strategy.unigen import UniGen
from sweetpea._internal.sampling_strategy.cmsgen import CMSGen
from sweetpea._internal.sampling_strategy.random import RandomGen
from sweetpea._internal.sampling_strategy.iterate_ilp import IterateILPGen
from sweetpea._internal.server import build_cnf
from sweetpea._internal.core.cnf import Var
from sweetpea._internal.argcheck import argcheck, make_islistof


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~ Top-Level functions ~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _experiments_to_tuples(experiments: List[dict],
                           keys: List[str]):
    """Converts a list of experiments into a list of lists of tuples, where
    each tuple represents a crossing in a given experiment.

    :param experiments:
        A list of experiments as :class:`dicts <dict>`. These are produced by
        calls to the synthesis function :func:`.synthesize_trials`.

    :returns:
        A list of lists of tuples of strings, where each sub-list corresponds
        to one of the ``experiments``, each tuple corresponds to a particular
        crossing, and each string is the simple surface name of a level.
    """
    tuple_lists: List[List[Tuple[str, ...]]] = []
    for experiment in experiments:
        tuple_lists.append(list(zip(*[experiment[key] for key in keys])))
    return tuple_lists


def simplify_experiments(experiments: List[Dict]) -> List[List[Tuple[str, ...]]]:
    return _experiments_to_tuples(experiments, list(experiments[0].keys()))


def experiments_to_tuples(block: Block,
                          experiments: List[dict]):
    return _experiments_to_tuples(experiments, [cast(str, f.name) for f in __filter_hidden(block.design)])


def __filter_hidden(design: List[Factor]) -> List[Factor]:
    return list(filter(lambda f: not isinstance(f.name, HiddenName), design))


def __filter_hidden_keys(d: dict) -> dict:
    return {name: d[name] for name in filter(lambda name: not isinstance(name, HiddenName), d.keys())}


def print_experiments(block: Block, experiments: List[dict]):
    """Displays the generated experiments in a human-friendly form.

    :param block:
        An experimental description as a :class:`.Block`.

    :param experiments:
        A list of experiments as :class:`dicts <dict>`. These are produced by
        calls to synthesis function :func:`.synthesize_trials`.
    """
    nested_assignment_strs = [list(map(lambda l: cast(str, f.name) + " " + str(l.name), f.levels))
                              for f in __filter_hidden(block.design)]
    column_widths = list(map(lambda l: max(list(map(len, l))), nested_assignment_strs))

    format_str = reduce(lambda a, b: a + '{{:<{}}} | '.format(b), column_widths, '')[:-3] + '\n'

    print('\n{} trial sequences found.\n'.format(len(experiments)))
    for idx, e in enumerate(experiments):
        print('Experiment {}:'.format(idx))
        strs = [list(map(lambda v: name + " " + str(v), values)) for (name, values) in e.items()]
        transposed = list(map(list, zip(*strs)))
        print(reduce(lambda a, b: a + format_str.format(*b), transposed, ''))


def tabulate_experiments(block: Block = None,
                         experiments: List[Dict] = None,
                         factors: Optional[List[Factor]] = None,
                         trials: Optional[List[int]] = None):
    """Tabulates and prints the given experiments in a human-friendly form.
    Outputs a table that shows the absolute and relative frequencies of
    combinations of factor levels.

    :param experiments:
        A list of experiments as :class:`dicts <dict>`. These are produced by
        a call to :func:`.synthesize_trials`.

    :param factors:
        An optional, though practically needed :class:`list` of :class:`Factors <.Factor>`.
        This list selects the factors of interest that are subsets of
        the design factors. More precisely, the names of these factors
        must be a subset of the design factor's names. (Given that :func:`.syntheseize_trials`
        reports results in terms of factor and level names, the design's factor and
        level objects no longer matter).

    :param trials:
        An optional :class:`list` of :class:`ints <int>` that specifies the indices of trials to
        be tabulated with the same indices being applied for every sequence. The paramater's
        default is to include all trials in the tabulation.
    """
    if factors is None and block is None:
        raise RuntimeError("tabulate_experiments: expected a `block` or `factors` argument")

    if factors is None:
        if (not isinstance(block, MultiCrossBlockRepeat)) or len(block.crossings) != 1:
            raise RuntimeError("tabulate_experiments: expected block with one crossing")
        factors = block.crossings[0]

    if experiments is None:
        raise RuntimeError("tabulate_experiments: need experiments")

    for exp_idx, e in enumerate(experiments):
        tabulation: Dict[str, List[str]] = dict()
        frequency_list = list()
        proportion_list = list()
        levels: List[List[str]] = list()

        if trials is None:
            trials = list(range(0, len(e[list(e.keys())[0]])))

        num_trials = len(trials)

        # initialize table
        for f in factors:
            tabulation[cast(str, f.name)] = list()
            factor_levels: List[str] = list()
            for l in f.levels:
                factor_levels.append(l.name)
            levels.append(factor_levels)

        max_combinations = 0
        # Each `element` is an n-tuple (s1, s2, ..., sn) where n is the number
        # of levels and each element is a level name.
        for element in product(*levels):
            max_combinations += 1

            # add factor combination
            for idx, factor_name in enumerate(tabulation.keys()):
                tabulation[factor_name].append(element[idx])

            # compute frequency
            frequency = 0
            for trial in trials:
                valid_condition = True
                for idx, factor in enumerate(tabulation.keys()):
                    if e[factor][trial] != element[idx]:
                        valid_condition = False
                        break
                if valid_condition:
                    frequency += 1

            proportion = frequency / num_trials

            frequency_list.append(str(frequency))
            proportion_list.append(str(proportion * 100) + '%')

        tabulation["frequency"] = frequency_list
        tabulation["proportion"] = proportion_list

        frequency_factor = Factor("frequency", list(set(frequency_list)))
        proportion_factor = Factor("proportion", list(set(proportion_list)))

        design = list()
        for f in factors:
            design.append(f)
        design.append(frequency_factor)
        design.append(proportion_factor)

        # print tabulation
        nested_assignment_strs = [list(map(lambda l: cast(str, f.name) + " " + l.name, f.levels)) for f
                                  in design]
        column_widths = list(map(lambda l: max(list(map(len, l))), nested_assignment_strs))

        format_str = reduce(lambda a, b: a + '{{:<{}}} | '.format(b), column_widths, '')[:-3] + '\n'

        print('Experiment {}:'.format(exp_idx))
        strs = [list(map(lambda v: name + " " + v, values)) for (name, values) in tabulation.items()]
        transposed = list(map(list, zip(*strs)))
        print(reduce(lambda a, b: a + format_str.format(*b), transposed, ''))


def _experiments_to_csv(experiments: List[dict],
                        csv_columns: List[str],
                        file_prefix: str = "experiment"):
    """Exports a list of experiments to CSV files. Each experiment will be
    saved to a separate ``.csv`` file.

    :param experiments:
        A list of experiments as :class:`dicts <dict>`. These are produced by
        a call to :func:`.synthesize_trials`.

    :param file_prefix:
        A prefix to attach to each output ``.csv`` file.
    """
    for idx, experiment in enumerate(experiments):

        dict = experiment
        num_rows = len(dict[csv_columns[0]])

        csv_file = file_prefix + "_" + str(idx) + ".csv"
        try:
            with open(csv_file, 'w') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerow(csv_columns)
                for row_idx in range(num_rows):
                    row = list()
                    for column in csv_columns:
                        row.append(dict[column][row_idx])
                    writer.writerow(row)
        except IOError:
            print("I/O error")


def save_experiments_csv(block: Block,
                         experiments: List[dict],
                         file_prefix: str = "experiment"):
    return _experiments_to_csv(experiments, [cast(str, f.name) for f in __filter_hidden(block.design)], file_prefix)


def synthesize_trials(block: Block,
                      samples: int = 10,
                      sampling_strategy=IterateGen
                      ) -> List[dict]:
    """Given an experiment described with a :class:`.Block`, randomly generates
    multiple sets of trials for that experiment.

    The number of trials in each set is determined by the experiment's
    crossing. Each trial is a combination of levels, subject to constraints
    imposed both implicitly and explicitly in the experimental description.

    This function should be used on the :class:`.Block` produced by either
    :func:`.fully_cross_block` or :func:`.multiple_cross_block`. Using that
    :class:`.Block`, :func:`.synthesize_trials` will produce a single cohesive
    CNF formula that will be solved with Unigen. The result of this solution is
    then decoded into something that is both human-readable and compatible with
    `PsyNeuLink <https://princetonuniversity.github.io/PsyNeuLink/>`_.

    :param block:
        An experimental description as a :class:`.Block`.

    :param samples:
        The number of trial sets to generate. For example, a value of ``1``
        produces a :class:`list` of length ``1``, which contains a single set
        of trials with a random ordering of the crossings that satisfies the
        given constraints.

        Default is ``10``.

    :param sampling_strategy:
        The strategy to use for trial generation. The default is
        :class:`.NonUniformGen`.

    :returns:
        A :class:`list` of trial sets. Each set is represented as a
        :class:`dictionary <dict>` mapping each factor name to a list of
        levels, where each such list contains to one level per trial.
    """

    def starting(who: Any) -> None:
        nonlocal samples
        print(f"Sampling {samples} trial sequences using {who}.")

    if isinstance(sampling_strategy, type):
        assert issubclass(sampling_strategy, Gen)
        starting(sampling_strategy.class_name())
        sampling_result = sampling_strategy.sample(block, samples)
    else:
        starting(sampling_strategy)
        sampling_result = sampling_strategy.sample_object(block, samples)

    trialss = list(map(lambda e: __filter_hidden_keys(block.add_implied_levels(e)),
                       sampling_result.samples))

    if os.getenv("SWEETPEA_CHECK_SYNTHESIZED"):
        for trials in trialss:
            mismatches = sample_mismatch_experiment(block, trials)
            if mismatches:
                print_experiments(block, [trials])
                print(mismatches)
                raise RuntimeError("synthesized trials has mismatches")

    return trialss


def sample_mismatch_experiment(block: Block, sample: dict) -> dict:
    """Given an experiment described with a :class:`.Block`, tests if :class:`list`
    of trials meets the factors, constraints and crossings of the described experiment.


    This function should be used on the :class:`.Block` produced by either
    :func:`.fully_cross_block` or :func:`.multiple_cross_block`. Using that
    :class:`.Block`, :func:`.synthesize_trials`.

    :param block:
        An experimental description as a :class:`.Block`.

    :param sample:
        A sample in the form of a :class:`list`.

    :returns:
        A :class:`dict` describing the mismatches. The entries of the dictionary lists the
        mismatches in the categories factors, constraints and crossings
    """
    res = {}
    for key in sample:
        if len(sample[key]) != block.trials_per_sample():
            res['trial_count'] = [key, len(sample[key]), block.trials_per_sample()]
    if not res:
        factor_errors = block.sample_mismatch_factors(sample)
        if factor_errors:
            res['factors'] = factor_errors
        constraint_errors = block.sample_mismatch_constraints(sample)
        if constraint_errors:
            res['constraints'] = constraint_errors
        crossing_errors = block.sample_mismatch_crossing(sample)
        if crossing_errors:
            res['crossings'] = crossing_errors
    return res


# TODO: This function isn't called anywhere, so it should be removed.
def save_cnf(block: Block, filename: str):
    """Generates a CNF formula from a :class:`.Block` and then writes that CNF
    formula to the indicated file in the Unigen-specific DIMACS format.

    :param block:
        A description of a CNF formula as a :class:`.Block`.

    :param filename:
        The name of the file to write the CNF formula to.
    """
    cnf_str = __generate_cnf(block)
    with open(filename, 'w') as f:
        f.write(cnf_str)


# ~~~~~~~~~~ Helper functions ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# TODO: This should be a method in Block.
def __generate_cnf(block: Block) -> str:
    """Converts a :class:`.Block` into a CNF formula and renders that CNF
    formula in the Unigen-specific DIMACS format.

    :param block:
        A description of a CNF formula as a :class:`.Block`.

    :returns:
        The given :class:`.Block` rendered as a Unigen-specific
        DIMACS-formatted string.
    """
    cnf = build_cnf(block)
    return cnf.as_unigen_string(sampled_variables=[Var(n) for n in block.support_variables()])



import pytest

from sweetpea import *

number = Factor("number", [1.0, 2.0, 3.0])
letter = Factor("letter", ["a", "b", "c"])

@pytest.mark.parametrize('index_and_solutions',
                         [[0, 2],
                          [-1, 2],
                          [-2, 2],
                          [100, 0],
                          [-100, 0]])
@pytest.mark.parametrize('strategy', [IterateSATGen, RandomGen])
def test_pin(index_and_solutions, strategy):
    index = index_and_solutions[0]
    solutions = index_and_solutions[1]
    block = CrossBlock([number], [number], [Pin(index, number[1.0])])
    experiments = synthesize_trials(block, 10, strategy)
    assert len(experiments) == solutions
    for e in experiments:
        trial_no = index if index >= 0 else len(e['number']) + index
        assert e['number'][trial_no] == 1.0

    block = CrossBlock([letter], [letter], [Pin(index, letter["b"])])
    experiments = synthesize_trials(block, 10, strategy)
    assert len(experiments) == solutions
    for e in experiments:
        trial_no = index if index >= 0 else len(e['letter']) + index
        assert e['letter'][trial_no] == "b"

    side = Factor("side", ["left", "right"])
    second_index = index + 3 if index > 0 else index - 3
    block = CrossBlock([letter, side], [letter, side], [Pin(index, (letter, "b")),
                                                        Pin(second_index, (letter, "b"))])
    experiments = synthesize_trials(block, 100, strategy)
    assert len(experiments) == solutions * 24
    for e in experiments:
        trial_no = index if index >= 0 else len(e['letter']) + index
        assert e['letter'][trial_no] == "b"
    
    unity = Factor("unity", [DerivedLevel("one", WithinTrial(lambda n: n == 1.0, [number])),
                             ElseLevel("many")])
    block = CrossBlock([number, unity], [number], [Pin(index, unity["many"])])
    experiments = synthesize_trials(block, 100, strategy)
    assert len(experiments) == solutions * 2
    for e in experiments:
        trial_no = index if index >= 0 else len(e['number']) + index
        assert e['number'][trial_no] != 1.0

    mood = Factor("mood", [DerivedLevel("up", Transition(lambda ns: ns[0] > ns[-1], [number])),
                           DerivedLevel("down", Transition(lambda ns: ns[0] < ns[-1], [number]))])
    t_index = index if index < 0 else index+1
    block = CrossBlock([number, mood], [number], [Pin(t_index, mood["up"])])
    experiments = synthesize_trials(block, 10, strategy)
    for e in experiments:
        trial_no = index if index >= 0 else len(e['number']) + index - 1
        e['number'][trial_no+1] > e['number'][trial_no]

@pytest.mark.parametrize('index_and_solutions',
                         [[0, 4, 12],
                          [-1, 4, 12],
                          [-2, 4, 12],
                          [100, 0, 0],
                          [-100, 0, 0]])
@pytest.mark.parametrize('strategy', [IterateSATGen, RandomGen])
def test_pin_repeats(index_and_solutions, strategy):
    index = index_and_solutions[0]
    solutions = index_and_solutions[1]
    repet_solutions = index_and_solutions[2]
 
    block = CrossBlock([number], [number], [Pin(index, number[1.0])])
    repet = Repeat(block, [MinimumTrials(6)])
    experiments = synthesize_trials(repet, 20, strategy)
    assert len(experiments) == solutions
    for e in experiments:
        trial_no = index if index >= 0 else len(e['number']) + index
        assert e['number'][trial_no] == 1.0
        trial_no = index + 3 if index >= 0 else len(e['number']) - 3 + index
        assert e['number'][trial_no] == 1.0

    block = CrossBlock([number], [number], [])
    repet = Repeat(block, [MinimumTrials(6), Pin(index, number[1.0])])
    experiments = synthesize_trials(repet, 20, strategy)
    assert len(experiments) == repet_solutions
    for e in experiments:
        trial_no = index if index >= 0 else len(e['number']) + index
        assert e['number'][trial_no] == 1.0

        
